"""Plot density mesh module"""
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import numpy as np
from arnica.solvers_2d.radiation import filter_stupid_characters


def heat_map_mesh(x_crd, y_crd, z_crd, show=False, save=False, view_axes='xr'):
    """ heat map plot of skin """

    plt.rcParams['image.cmap'] = 'Blues'

    def get_bins(x_crd, y_crd):
        """Computes equal bins from aspect ratio"""
        l_y = np.abs(np.max(y_crd) - np.min(y_crd))
        l_x = np.abs(np.max(x_crd) - np.min(x_crd))
        ratio = l_y / l_x
        return 500., 500. * ratio

    def plot_skin(tuple_data, tuple_labels):
        """Plot skin as hist2d"""
        x_crd, y_crd = tuple_data
        x_label, y_label = tuple_labels

        fig = plt.figure()
        ax_ = fig.add_subplot(111)

        title = "%s, %s" % tuple_labels
        ax_.set_title(title)
        ax_.hist2d(x_crd, y_crd,
                   bins=get_bins(x_crd, y_crd),
                   norm=LogNorm())
        ax_.set_xlabel("%s" % x_label)
        ax_.set_ylabel("%s" % y_label)
        ax_.set_aspect(aspect=1)
        if save:
            name = filter_stupid_characters(title)
            plt.savefig('%s' % name)

    radius = np.hypot(y_crd, z_crd)


    if view_axes == 'xr':
        plot_skin((x_crd, radius), ('$x$', '$r$'))
    elif view_axes == 'xz':
        plot_skin((x_crd, z_crd), ('$y$', '$z$'))
    elif view_axes == 'xy':
        plot_skin((x_crd, y_crd), ('$x$', '$y$'))
    elif view_axes == 'zy':
        plot_skin((z_crd, y_crd), ('$z$', '$y$'))
    else:
        plot_skin((x_crd, radius), ('$x$', '$r$'))
        plot_skin((x_crd, z_crd), ('$y$', '$z$'))
        plot_skin((x_crd, y_crd), ('$x$', '$y$'))
        plot_skin((z_crd, y_crd), ('$z$', '$y$'))

    if show:
        plt.show()


def scatter_plot_mesh(x_crd, y_crd, z_crd, axisym=False, show=False):
    """ scatter plot of skin """

    if axisym:
        radius = np.sqrt(np.square(x_crd) + np.square(y_crd))
        theta = np.arctan2(z_crd, y_crd)

        plt.figure()
        plt.title(r"x, theta")
        plt.scatter(x_crd, theta, edgecolors='none', alpha=0.25)
        plt.xlabel("x")
        plt.xlabel("theta")

        plt.figure()
        plt.title("x, r")
        plt.scatter(x_crd, radius, edgecolors='none', alpha=0.25)
        plt.xlabel("x")
        plt.xlabel("x")

        plt.figure()
        plt.title("r, theta")
        plt.scatter(radius, theta, edgecolors='none', alpha=0.25)
        plt.xlabel("r")
        plt.xlabel("theta")
    else:
        plt.figure()
        plt.title("z, y")
        plt.scatter(z_crd, y_crd, edgecolors='none', alpha=0.25)
        plt.xlabel("z")
        plt.xlabel("y")

        plt.figure()
        plt.title("x, y")
        plt.scatter(x_crd, y_crd, edgecolors='none', alpha=0.25)
        plt.xlabel("x")
        plt.xlabel("y")

        plt.figure()
        plt.title("x, z")
        plt.scatter(x_crd, z_crd, edgecolors='none', alpha=0.25)
        plt.xlabel("x")
        plt.xlabel("z")

    if show:
        plt.show()



"""
*Plot graphs from 1D average array with or without its\
 confidence interval. Rotate the graph from 90 deg.*
"""
import matplotlib.pyplot as plt

def plot_ave_with_interval(x_arr,
                           average,
                           profile="average-interval",
                           upper=None,
                           lower=None,
                           **kw_args):
    """
    *Plot average profile with or without confidence interval*

    :param x_arr: Array of float of x-axis
    :type x_arr: np.array
    :param average: Array of float of average curve
    :type average: np.array
    :param profile: Plot type (average-interval, average, integral)
    :type profile: str
    :param upper: Array of float of upper interval values
    :type upper: np.array
    :param lower: Array of float of lower interval values
    :type upper: np.array

    Optional Keyword Args:

    :param x_label: Label for x-axe
    :type x_label: str
    :param y_label: Label for y-axe
    :type y_label: str
    :param style: Style of the axes - plain or sci
    :type style: str

    :return:

        - **plt** - Matplotlib.pyplot object
    """
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals

    fig = plt.figure(figsize=(5.5, 5))
    fig.subplots_adjust(left=0.15, right=0.97, bottom=0.15,
                        top=0.9, wspace=0.27)

    axe = fig.add_subplot(111)

    x_label = kw_args.get('x_label')
    if x_label is None:
        x_label = "x"

    y_label = kw_args.get('y_label')
    if y_label is None:
        y_label = "y"

    axe.set_ylabel(y_label)
    axe.set_xlabel(x_label)

    if "average" in profile:
        axe.plot(average, x_arr, c='k', lw=0.5)

    if "interval" in profile and upper is not None and lower is not None:
        axe.fill_betweenx(x_arr, upper, lower, facecolor='red', alpha=0.4)

    axe.grid(which='major', color='gray', linestyle='--',
             dashes=(8, 12), linewidth=0.5)

    # Style = 'plain' | 'sci'
    style = kw_args.get('style')
    if style is None:
        style = 'plain'
    axe.ticklabel_format(axis='both', style=style, scilimits=(0, 0))

    return plt



""" Module to plot a field on cartesian mesh through a triangular mesh """

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.tri as tri
from mpl_toolkits.axes_grid1 import make_axes_locatable

def get_connectivity(i, j, shape):
    """
    *Generate connectivities*

    :param i: Index of the cell on u-axis
    :type i: int
    :param j: Index of the cell on v-axis
    :type j: int
    :param shape: Shape of the shell

    :returns:

        - **connectivity** - Array of shape (t, 3) of triangle point indexes
    """
    connectivity = []
    connectivity.append([i*shape[1] + j,
                         i*shape[1] + (j+1),
                         shape[0] * shape[1] + i*(shape[1]-1) + j])
    connectivity.append([i*shape[1] + (j+1),
                         (i+1)*shape[1] + (j+1),
                         shape[0] * shape[1] + i*(shape[1] - 1) + j])
    connectivity.append([(i+1)*shape[1] + (j+1),
                         (i+1)*shape[1] + j,
                         shape[0] * shape[1] + i*(shape[1] - 1) + j])
    connectivity.append([(i+1)*shape[1] + j,
                         i*shape[1] + j,
                         shape[0] * shape[1]+ i*(shape[1] - 1) + j])

    return np.asarray(connectivity)

def quad2tri(grid, field):
    """
    *Divide a cartesian grid into an extended triangular grid*

    :param grid: Array of shape (n_u, n_v, 2) of cartesian coordinates
    :param field Array of shape (n_u, n_v) of field values

    :returns:

        - **triangulation** - Matplotlib.tri.Triangulation object containing \
                              extended grid coordinates and connectivity array
        - **field_at_node** - Array of extended field values stored at the node
    """

    shape = grid.shape[0:2]
    grid_center = np.ones((shape[0] - 1, shape[1] - 1, 2))
    field_center = np.ones((shape[0] - 1, shape[1] - 1))

    connectivity = np.empty((0, 3), dtype=int)
    for i in range(shape[0] - 1):
        for j in range(shape[1] - 1):
            grid_center[i, j] = np.mean(grid[i:i+2, j:j+2].reshape(4, 2), axis=0)
            connectivity = np.append(connectivity, get_connectivity(i, j, shape), axis=0)
            field_center[i, j] = np.mean(field[i:i+2, j:j+2].reshape(4))

    coord = np.stack((np.append(np.ravel(grid[:, :, 0]), np.ravel(grid_center[:, :, 0])),
                      np.append(np.ravel(grid[:, :, 1]), np.ravel(grid_center[:, :, 1])))).T
    field_at_node = np.append(np.ravel(field), np.ravel(field_center))
    #field_at_cell = field_at_node[connectivity].mean(axis=-1)

    triangulation = tri.Triangulation(coord[:, 0], coord[:, 1], connectivity)

    return triangulation, field_at_node

def plot_quad2tri(grid, title, field, shading=True):
    r"""\
    *Generate a plot of a field on a x,y shell cartesian grid*

    The cartesian grid is converted into a triangle grid with connectivities.
    The field is extended by interpolation on the new points created.
    The plot is generated from the points coordinates extended and the connectivity, \
    using tripcolor of matplotlib.

    If shading is True : using 'gouraud' from field value defined at nodes.
    If shading is False : using 'flat' from mean field value defined at the nodes or \
                          field value defined at the cell.

    :param grid: Array of shape (n_u, n_v, 2) of cartesian coordinates
    :param title: Title of the plot
    :type title: str
    :param field: Array of shape (n_u, n_v) of field values
    :param shading: Define if plot is shaded of not
    :type shading: bool

    :returns:

        - **plt** - Matplotlib object containing the graph
    ::
            1 quad              4 tri
          O---------O        O---------O
          |         |        | \     / |
          |         |        |  \   /  |
          |         |        |   \ /   |
          |         |  --->  |    O    |
          |         |        |   / \   |
          |         |        |  /   \  |
          |         |        | /     \ |
          O---------O        O---------O

    """

    triangulation, field_at_node = quad2tri(grid, field)

    fig1, ax1 = plt.subplots()
    ax1.set_aspect('equal')
    if shading:
        shading = 'gouraud'
    else:
        shading = 'flat'

    tpc = ax1.tripcolor(triangulation,
                        field_at_node,
                        shading=shading)

    divider = make_axes_locatable(ax1)
    cax = divider.append_axes("right", size="5%", pad=0.1)

    fig1.colorbar(tpc, cax=cax)
    ax1.set_title(title)
    ax1.set_xlabel('z (m)')
    ax1.set_ylabel('y (m)')

    return plt



""" axishell module
to create x-axisymmetric shells for scientific computations
"""

import numpy as np
from arnica.utils.shell import Shell

class CartShell(Shell):
    #pylint: disable=too-many-instance-attributes
    r"""\
    *Base class for cartesian computationnal shells*

    CartShell class is based on Shell class, inheriting of Shell methods.
    At initialization, the shell is builded an an cartesian shell parallel to z.

    Some attributes are based on **u** and **v**, defined as the curvilinear\
    longitudinal (x/y) abscissa and curvilinear transversal (z) abscissa\
    respectively.

    ::
                   (x_n, y_n)
          ________X________
          |       |       |            Cartesian system
          |       |       |            Example longi/u <=> x
          |       |       |
          |       |       |            x - longi/u
          |    <----->    |            ^
          |       |       |            |
          |       |       |            |
          |       |       |            O-----> z - transvers/v
          |_______X_______|           y
                   (x_0, y_0)

    :param n_transvers: Number of azimuthal shell points
    :type n_transvers: int
    :param n_longi: Number of longitudinal shell points
    :type n_longi: int
    :param z_range: Extrusion range in transvers direction
    :type z_range: float
    :param ctrl_pts_x: Array of dim (n,) of x-coordinates of the points defining the spline
    :param ctrl_pts_y: Array of dim (n,) of y-coordinates of the points defining the spline

    Optionals arguments :
    :param z_min: Minimum transversal value of the shell
    :type z_min: float

    Public attributes :

        - **shape** - Shape of the shell : (n_transvers, n_longi)
        - **width_matrix** - Dict containing thickness matrices of shape 'shape'
        - **matrix** - Dict containing fields of shape 'shape' - To deprecate

    Private attributes :

        - **_xyz**
        - **_rad**
        - **_theta**
        - **_n_x**
        - **_n_r**
        - **_n_y**
        - **_n_z**
        - **_du**
        - **_dv**
        - **_abs_curv**
        - **_dwu**
        - **_dwv**
        - **_surf**

    """

    def __init__(self,
                 n_trans,
                 n_longi,
                 z_range,
                 ctrl_pts_x,
                 ctrl_pts_y,
                 z_min=None):
        #pylint: disable=too-many-arguments
        """
        *Initialize a CartShell object*
        """

        Shell.__init__(self, n_trans, n_longi)

        self.ctrl_pts_x = ctrl_pts_x
        self.ctrl_pts_y = ctrl_pts_y
        self.z_range = z_range
        self.z_min = z_min

        self._build_shell()

    def _build_shell(self):
        #pylint: disable=arguments-differ
        """
        *Build shell from geometric features*

            - Construct a spline used as base for extrusion\
              from control points : tck
            - Discretise the spline : shell_crest
            - Compute normal vectors for the 1D shell_crest
            - Compute r,n_x,n_r-components for 2D shell
            - Compute theta-components for 2D shell
            - Compute xyz,n_y,n_z-components for 2D shell
        """

        # Construct Shell Crest
        shell_crest = self._compute_shell_crest(self.ctrl_pts_x, self.ctrl_pts_y)

        # Compute xyz matrix of shape (n_transvers, n_longi, 3)
        tmp_x = np.tile(shell_crest[0], (self.shape[0], 1))
        tmp_y = np.tile(shell_crest[1], (self.shape[0], 1))

        trans_z = 0.
        if self.z_min is not None:
            trans_z = 0.5 * self.z_range + self.z_min
        min_z = trans_z - 0.5 * self.z_range
        max_z = trans_z + 0.5 * self.z_range
        tmp_z = np.transpose(np.tile(np.linspace(min_z,
                                                 max_z,
                                                 num=self.shape[0]),
                                     (self.shape[1], 1)))
        self._xyz = np.stack((tmp_x, tmp_y, tmp_z), axis=-1)

        # Compute radius and azimuth matrix of shape (n_transvers, n_longi)
        self._rad = np.sqrt(np.take(self._xyz, 1, -1) ** 2
                            + np.take(self._xyz, 2, -1) ** 2)
        self._theta = np.arctan2(np.take(self._xyz, 2, -1),
                                 np.take(self._xyz, 1, -1))

        # Compute x,y,z,r-normal matrix of shape (n_transvers, n_longi)
        xy_nml_1d = self._compute_shellcrest_nml(shell_crest)
        self._n_x = np.tile(xy_nml_1d[0], (self.shape[0], 1))
        self._n_y = np.tile(xy_nml_1d[1], (self.shape[0], 1))
        self._n_z = np.zeros(self.shape)
        self._n_r = self._n_y * np.cos(self._theta)

        # Compute du,dv matrix of shape (n_transvers, n_longi)
        self._du = np.pad(np.sqrt(np.diff(self._xyz[:, :, 0], axis=1) ** 2
                                  + np.diff(self._xyz[:, :, 1], axis=1) ** 2),
                          ((0, 0), (1, 0)),
                          'edge')
        self._dv = np.pad(np.diff(self._xyz[:, :, 2], axis=0),
                          ((1, 0), (0, 0)),
                          'edge')
        # Compute abs_curv array of shape (n_longi,)
        self._abs_curv = np.cumsum(np.take(self._du, 0, 0)) - self._du[0, 0]

        # Compute weight intervals in u and v directions array of shape (n_transvers, n_longi)
        self._dwu = self._du.copy()
        self._dwu[:, (0, -1)] /= 2
        self._dwv = self._dv.copy()
        self._dwv[(0, -1), :] /= 2

        # Compute surface nodoes array of shape (n_transvers, n_longi)
        self._surf = np.multiply(self._dwu, self._dwv)



"""
==================
Physical utilities
==================

These physical tools are helpers around CFD-related problems.

- **solid_material** is a class to store solid properties for CHT problems.
- **thermodyn_properties** is a set of tools for propertiesand correlations used in CHT problems
- **wall_thermal_equilibrium** compute the thermal equilibrium for a 2-layer wall (Metal/ceramic)
- **yk_from_phi** compute the mass fraction set according to equivalence ratio.

"""


from .solid_material import *
from .thermodyn_properties import *
from .wall_thermal_equilibrium import *
from .yk_from_phi import *




""" SORT AND SAMPLE POINTS FROM CURVE """

import time
import numpy as np
from scipy import spatial

def get_neighbor(kdtree, point, list_points):
    """
    *Find the closiest neighbor that has not already been found*

    From the kdtree, find the two closiest points of 'point', the 1st\
    being itself.
    If the second closiest point has already been found, then the\
    number of researched points is increased until finding a point that\
    has not already been found.

    :param kdtree: KDTree of all points.
    :param point: Array of dim (k,) of point float coordinates.
    :type point: np.array
    :param list_points: List of integer indexes

    :return:

        - **index** - Index (int) of the unfound closiest point.
        - **dist** - Distance (float) of the closiest point from point.
    """

    k_neighbors = 2
    dists, indexes = kdtree.query(point, k=k_neighbors)
    subindexes = np.setdiff1d(indexes, list_points, assume_unique=True)

#    while indexes[-1] in list_points:
    while len(subindexes) == 0:
        k_neighbors += 1
        dists, indexes = kdtree.query(point, k=k_neighbors)
        subindexes = np.setdiff1d(indexes, list_points, assume_unique=True)
        #if k_neighbors == 4000:
        #    print("k diverges")
        #    stop = True
        #    break

    index = subindexes[0]
    dist = dists[np.where(indexes == subindexes[0])[0][0]]

    return index, dist

def sort_points_by_dist(points_coor, starting_pt):
    #pylint: disable=too-many-locals
    r"""\
    *Reorder a point cloud by distances*
                                                   .
    From a starting fictive point, the first point 1 of the spline\
    iget_neighbors found, as the closiest one.       .
    From that point, the following points are obtained with get_neighbor()
    until having sorted all points.

    ::
            ................                   ................       r/y
        ...      2     8    ..             ...                 ..    ^
      ..                     .           ..                     .    |
      .                      .           .                      .    |      x
      .      ................     ===>   .      ................     o----->
      .     .                            .     .
       .     ...................          .     ...................
       ..                                 ..
          ......................             ......................
            1         3          X                         8    321
                                  starting_pt

    :param points_coor: Array of dim (n,k) of points coordinates
    :type points_coor: np.array
    :param starting_pt: Array of dim (k,) of starting pt coordinates
    :type starting_pt: np.array

    :returns:

        - **ordered_indexes** - Array of dim (n,) of indexes
        - **ordered_dists** - Array of dim (n,) of floats
    """

    kdtree = spatial.cKDTree(points_coor) #pylint: disable=not-callable

    list_points = []
    list_dists = []

    _, idx = kdtree.query(starting_pt, k=1)
    list_points.append(idx)
    list_dists.append(0)

    progress = -1
    time_i = time.time()
    while len(list_points) < len(points_coor):

        index, dist = get_neighbor(kdtree,
                                   points_coor[list_points[-1]],
                                   list_points)

        list_dists.append(dist)
        list_points.append(index)

        new_progress = int(len(list_points)/len(points_coor) * 100)
        if int(new_progress / 10) > int(progress / 10):
            progress = new_progress
            delta_time = time.time() - time_i
            print("Sorting process... {:d}% : {:.3f}s".format(progress, delta_time), end="\r")

    print()
    ordered_indexes = np.asarray(list_points)
    ordered_dists = np.asarray(list_dists)

    return ordered_indexes, ordered_dists

def sample_arrays_by_dist_interval(dists, samples_res, *args):
    r"""\
    *Sample data and optional args at each sample_res along data*

    From an array containing the distance between sorted points, \
    and from a sample resolution defined by the distance between\
    two samples, the function picks up the indexes corresponding\
    to a sample, and returns an array of sampled distances and\
    arrays of sampled arguments from those indexes.

    __ = samples_res

    ::
            ................                   .__.__.__.__.__.
        ...                 ..             .__                 ..
      ..                     .           ..                     |
      .                      .           |                      .
      .      ................     ===>   |      .__.__.__.__.__.
      .     .                            .     .
       .     ...................          .     .__.__.__.__.__.__.
       ..                                 ..
          ......................           9 .__.__.__.__.__.__.__.
                      9      321                            3  2  1


    :param dists: Array of dim (n,) of floats
    :type dists: np.array
    :param samples_res: Sample resolution for sampling
    :type samples_res: float
    :param args: Tuple de data of dim (n,)
    """

    cum_dists = np.cumsum(dists, axis=0)
    sum_dists = cum_dists[-1]
    indexes = [0]

    for dist in np.arange(samples_res, sum_dists, samples_res):
        next_index = np.where((cum_dists >= dist) & (cum_dists < dist + samples_res))[0]
        if next_index.size > 0:
            indexes.append(next_index[0])

    if indexes[-1] != len(dists) - 1:
        indexes[-1] = len(dists) - 1

    dists = dists[indexes]
    args_out = []
    for arg in args:
        args_out.append(arg[indexes])

    return dists, tuple(args_out)

def sample_points_from_cloud(points_coor,
                             starting_pt,
                             n_samples=None,
                             samples_res=None):
    """
    *Sort and sample unsorted curve*

    First the point coordinates are sorted by distances.
    Secondly compute sample resolution if not provided.
    Finally sample the ordered indexes points by distance.
    Returns the array of coordinates ordered and sampled.

    :param points_coor: Array of dim (n,k) of points coordinates
    :type skin_coor: np.array
    :param starting_pt: Array of dim (k,) of starting point coordinates
    :type starting_pt: np.array
    :param n_samples: Number of samples to extract
    :type n_samples: int
    :param sample_res: Resolution of the sampling
    :type sample_res: float

    :returns:

        - **skin_cyl** - Array of dim (n_samples, 3) of coordinates
    """

    # Sort the points and get samples of them.
    ordered_indexes, ordered_dists = sort_points_by_dist(points_coor, starting_pt)

    # Calculate optional variable from specified one.
    if samples_res is None:
        # Sample resolution : distance between two point samples.
        if n_samples is None:
            n_samples = 80
        samples_res = np.sum(ordered_dists) / n_samples

    _, (sampled_ordered_indexes,) = sample_arrays_by_dist_interval(ordered_dists,
                                                                   samples_res,
                                                                   ordered_indexes)

    return points_coor[sampled_ordered_indexes]



"""
Efficient implementation of unstructured mesh operations

Created Feb 8th, 2018 by C. Lapeyre (lapeyre@cerfacs.fr)
"""
import numpy as np
from scipy import sparse


class UnstructuredAdjacency():
    """Efficient scipy implementation of unstructured mesh adjacency ops

    The connectivity is stored in a sparse adjencency matrix A of shape
    (nnode, ncell). The gather operation on vector X (nnode) yields the
    scattered vector Y (ncell), and the scatter operation yields the
    filtered vector X' (nnode). This writes:
        Y  = 1/nvert . A . X
        X' = 1/bincount . A^t . Y
    where ^t is the transpose operation.

    The gatter-scatter operation resulting in filtering X can be performed
    efficiently by storing:
        F  = 1/bincount . A^t . 1/nvert . A
        X' = F . X
    """
    
    def __init__(self, connectivity):
        """Store connectivity as adjacency matrix, extract global parameters

        connectivity: a dictionary-like object with
            {number_of_vertex: connectivity}
          where:
            - number_of_vertex : an integer (number of vertex per element)
            - connectivity     : a list of node indices describing elements of
                                 length number_of_vertex * number_of_cells
          One entry is expected per element type. Ex:
              {3: [triangle_cells_node_list],
               4: [quad_cells_node_list]}
        """
        self._connectivity = connectivity
        concat = np.hstack(list(connectivity.values()))
        self.nnode = np.unique(concat).size
        min_node_number = concat.min()
        # Ensure node numbering starts at 0
        self._connectivity = {nvert: conn - min_node_number
                              for nvert, conn in self._connectivity.items()}
        self._bincount = np.bincount(concat - min_node_number)
        self._adjacency_matrix = self._make_adjacency()

    @property
    def ncell(self):
        """Total number of cells"""
        return sum(self.ncell_per_nvert.values())

    @property
    def ncell_per_nvert(self):
        """Dictionary of {nvert: ncell}

        For each type of element with nvert vertices, stores the nubmer of
        cells ncell
        """
        return {nvert: conn.size // nvert
                for nvert, conn in self._connectivity.items()}

    def _make_adjacency(self, weights=None):
        """Create adjacency matrix by reading connectivity"""
        if weights is None:
            weights = {nvert: np.ones(self._connectivity[nvert].shape)
                       for nvert in self._connectivity}
        adjacency = []
        for nvert in sorted(self._connectivity):
            conn = self._connectivity[nvert]
            ncell = self.ncell_per_nvert[nvert]
            indptr = np.arange(ncell + 1) * nvert
            adjacency.append(sparse.csr_matrix((weights[nvert], conn, indptr),
                                               shape=(ncell, self.nnode)))
        return sparse.vstack(adjacency)

    @property
    def _node2cell_matrix(self):
        """The (ncell, nnode) matrix"""
        node_weights = {nvert:
                        np.ones(self.ncell_per_nvert[nvert] * nvert) / nvert
                        for nvert in sorted(self.ncell_per_nvert)}
        return self._make_adjacency(weights=node_weights)

    @property
    def _cell2node_matrix(self):
        """The (nnode, ncell) matrix"""
        mat = self._make_adjacency().T
        weights = sparse.diags(1. / self._bincount)
        return weights * mat

    def get_node2cell(self):
    #def get_node2cell(self, node_vector):
        """Return the node2cell function"""
        return lambda node_vector: self._node2cell_matrix * node_vector

    def get_cell2node(self):
        """Return the cell2node function"""
        return lambda cell_vector: self._cell2node_matrix * cell_vector

    def get_filter(self, times=1):
        """Return the full gather + scatter filter operation

        If you need to perform the operation N times, you can use the
        times attribute.

        """
        gatherscatter = self._cell2node_matrix * self._node2cell_matrix

        def filt(node_vector):
            """Function to return by get_filter method"""
            for _ in range(times):
                node_vector = gatherscatter * node_vector
            return node_vector
        return filt



"""
=========
The utils
=========

These utils are helpers around CFD-related problems.

- **showy** is a matplotlib helper for using subplots with re-usable templates.
- **show_mat** is a matplotlib helper for fast matrix plotting with legend and axis naming.
- **cloud2cloud** is an inverse distance interpolator without connectivity.
- **directed_projection** is a projection of vectors clouds along their directions.
- **vector_actions** is a set of vector transformation helpers.
- **plot_density_mesh** is a mesh rendering tool using matplotlib hist2d.
- **axi_shell** is a 2D i-j structured mesh mapping axycylindrical splaine-based surfaces.
- **nparray2xmf** is a 1-2-3D i-j-k structured numpy datastructure dumping facility to XDMF format.

Untested - to be deleted :
--------------------------

- **unstructured_adjacency** *unstested* is the beginning of mesh handling using connectivity.
- **mesh_tools** *unstested* is a 2D mesh generation in numpy for solvers
- **datadict2file** was a dumping facility for dictionnary-like data. To be replaced by *hdfdict* or h5py-wrapper* packages.
- **timer_decorator** is a lightweight timer for functions. Better to use cProfile...

"""

from .axishell import *
from .cloud2cloud import *
from .directed_projection import *
from .nparray2xmf import *
from .vector_actions import *
from .showy import *
from .sample_curve_from_cloud import *
from .same_nob import *



""" Module to compute the directed projection
of a point to a surface along a direction

::

        ___
            \
             \
        x---->\
               \

OST : Seven nation Army (Westworld), R. Djawadi
"""

from collections import namedtuple
import warnings
from scipy import spatial
import numpy as np

BIG = 1e6

def projection_kdtree(points,
                      directions,
                      point_surface,
                      normal_surface,
                      **kwargs):
    """
    Project the n points following the direction on the m suface.

    :param points: Array of [p] points of shape (p,3)
    :param directions: Array of [p] direction vectors of shape (p,3)
    :param point_surface: Array of [n] surface nodes of shape (n,3)
    :param normal_surface: Array of [n] surface node normals of shape (n,3)
    :param neigbors: Number [k] of neigbors to take into account
    :type neigbors: int
    :param tol: Maximum distance beyond cyl dist with big set to BIG
    :type tol: float
    :param project: If True, first project points along normal.
    :type project: bool

    :returns:
        - **projected_pts** - "t" nparray of shape (n,3), projected on the surface
        - **indexes** - neigborhood of the points (n,k)
        - **cyl_dist** - cylindrical distance of p with each neighbor (n,k)

    ::

               < shp_dist >
        ______s___________t__________________
              |' .                       A
              |     '  .< cyl_dist >     .
              v                          .< surface_dist>
                     4                   .
                    /                    .
                   /                     .
                  p                      V

              align : alignment (pscal of two unit vectors, in [-1,1])

    Algorithm :
        - If project bool is True, first compute [p] projection from [p] points along the [p,1] \
          spherical closest node's normal. If False, projected_points = points.
        - Reduce computation to the [p,k] spherical closest nodes of the [p] projected_points
        - Compute the [p,k] cylindrical distances from the [p,k] closest nodes \
          to the [p] lines defined by the [p] projected points and the [p] direction vectors.
    """
    opts = {'neigbors': 1000, 'tol': 1000.0, 'project': True}
    if 'proj_axis' in kwargs:
        warnings.warn("'proj_axis' kwarg is deprecated. Please use boolean 'project' "
                      "kwarg instead : project = True if proj_axis = 'nml', false instead.")
        if kwargs['proj_axis'] == 'dir':
            opts['project'] = False
    for keyword in opts:
        if keyword in kwargs:
            opts[keyword] = kwargs[keyword]

    print("KDTree with  "
          + "\n-   Surface points | "
          + str(point_surface.shape[0])
          + "\n- Projected points | "
          + str(points.shape[0])
          + "\n-        Neighbors | "
          + str(opts['neigbors'])
          )

    kdtree = spatial.cKDTree(point_surface) # pylint: disable=not-callable
    if opts['project']:
        _, index = kdtree.query(points, k=1)
        projected_pts = project_points(
            points,
            normal_surface[index],
            point_surface[index][:, np.newaxis, :]).proj_pts
        projected_pts = np.squeeze(projected_pts, axis=1)
    else:
        projected_pts = points

    _, indexes = kdtree.query(projected_pts, k=opts["neigbors"])
    cyl_dists = compute_cyl_dist(
        projected_pts,
        directions,
        point_surface[indexes],
        normal_surface[indexes],
        opts["tol"])

    Proj_KDTree = namedtuple("Proj_KDTree", ["moved_pts", "indexes", "cyl_dists"])
    return Proj_KDTree(projected_pts, indexes, cyl_dists)

def compute_cyl_dist(points, directions, points_surf, normals_surf, tol):
    """
    *Compute cylindrical distances*

    For a i-number of points coordinates, compute the cylindrical distances\
    between the i,p-number of nodes with the i-number of axis.

    The array is then clipped according to the node normals and the direction\
    of the drills.

    :param points: Array of dim (i,3) of drill float coordinates
    :type points: np.array
    :param directions: Array of dim (i,3) of drill float axis components
    :type directions: np.array
    :param points_surf: Array of dim (i,p,3) of nodes float coordinates
    :type points_surf: np.array
    :param normals_surf: Array of dim (i,p,3) of nodes float normal components
    :type normals_surf: np.array
    :param params: Dict of parameter

    :returns:

        - **cyl_dists** - Array of dim (i,p) of float cylindrical distances
    """

    proj = project_points(points, directions, points_surf)
    cyl_dists = proj.rad_dists
    axi_dists = proj.axi_dists
    align = np.sum(np.multiply(normals_surf, directions[:, np.newaxis, :]), axis=-1)
    cyl_dists = np.where(align < 0.0,
                         cyl_dists,
                         BIG*np.ones(cyl_dists.shape))
    cyl_dists = np.where(np.absolute(axi_dists) < tol,
                         cyl_dists,
                         BIG*np.ones(cyl_dists.shape))

    return cyl_dists

def project_points(points_source, normals, points_target):
    r"""\
    *Compute the projected points, radial dists and axial dists*

    Compute projection from source points S of dim (k,) or (i,k),\
    on a plan defined by normals N of dim (k,), (i,k), (p,k) or (i,p,k),\
    and points T of dim (k,), (i,k), (p,k), (i,p,k).
    With :

        - **i** : Number of points to project
        - **p** : Number of points defining plans
        - **k** : Dimension of the domain

    ::

               S            T1          T2
                \          ⁄           ⁄
                 \       ⁄           ⁄
           axi_d1 \ ⁄\ ⁄ rad_d1    ⁄
                   \ ⁄           ⁄ rad_d2
                    \          ⁄
                     \       ⁄
               axi_d2 \ ⁄\ ⁄
                       \ ⁄
                        X proj_pt2
                         \
                          N

    S     : Points source
    N     : Normal
    T     : Points target
    axi_d : Axial distance of the point T projected on the axis Ax
    rad_d : Cylindrical or Radial distance between T and the axis Ax

               ------>   ->
    axi_dist = (T - S) . N

    -------------->   ->  ->
    projected_point = S + N * axi_dist

                    ->   -------------->
    rad_dist = norm(T - projected_point)

    :param points_source: Array of source points coordinates
    :type points_source: np.array
    :param proj_axis: Array of normal components defining projection plans
    :type proj_axis: np.array
    :param points_target: Array of points coordinates defining projection plans
    :type points_target: np.array

    :return:

        - **projected_points** - Array of shape points_target.shape of float coordinates
        - **axi_dists** - Array of shape points_target.shape[:-1] of float distances
        - **rad_dists** - Array of shape points_target.shape[:-1] of float distances
    """
    points_source = points_source[:, np.newaxis, :]
    normals = normals[:, np.newaxis, :]

    axi_dists = np.sum(np.multiply((points_target - points_source),
                                   normals),
                       axis=-1)
    projected_points = points_source + np.multiply(normals, axi_dists[:, :, np.newaxis])

    rad_dists = np.linalg.norm(points_target - projected_points, axis=-1)

    Proj = namedtuple('Proj', ['proj_pts', 'axi_dists', 'rad_dists'])

    return Proj(projected_points, axi_dists, rad_dists)



"""
lay_and_temp_manager.py

Functions which deal with layouts and templates
"""

from copy import deepcopy
from yaml import safe_load
from pkg_resources import resource_listdir, resource_string
import pandas as pd
from arnica.utils.string_manager import (string_minus_string,
                                         replace_stars_in_string)

__all__ = ["fetch_avbp_layouts", "fetch_avbp_templates", "decompact_template"]


def fetch_avbp_layouts():
    """It returns all the avbp layouts in a dictionary

    Output:
    -------
    avbp_layouts : nested object
    """

    avbp_layouts = {}
    layouts_files_list = resource_listdir('pyavbp.visu', 'avbp_layouts')

    for layout_file_name in layouts_files_list:
        layout_string = resource_string('pyavbp.visu', ("avbp_layouts/"
                                                        + layout_file_name))
        layout = safe_load(layout_string)
        avbp_layouts[layout["title"]] = layout

    return avbp_layouts


def fetch_avbp_templates():
    """It returns all the avbp templates in a dictionary

    Output:
    -------
    avbp_templates : nested object
    """

    avbp_templates = {}
    templates_files_list = resource_listdir('pyavbp.visu', 'avbp_templates')

    for template_file_name in templates_files_list:
        template_string = resource_string('pyavbp.visu', ("avbp_templates/"
                                                          + template_file_name))
        template = safe_load(template_string)
        avbp_templates[template["title"]] = template

    return avbp_templates


def decompact_template(template, data):
    """It decompacts the provided template in function of the provided
    data which are in the form of a key-value object and returns it

    Input:
    ------
    template : nested object
    data : key-value object

    Output:
    -------
    layout : nested object
    """

    if not isinstance(data, pd.DataFrame):
        dataframe = pd.DataFrame.from_dict(data)
    else:
        dataframe = data

    layout = deepcopy(template)
    layout['graphs'] = []

    for graph_template in template['graphs']:
        layout['graphs'] += decompact_graph_template(graph_template,
                                                     dataframe)

    return layout


def decompact_graph_template(graph_template, dataframe):
    """It returns the layouts of graphs associated to the provided
    template in function of the provided data frame

    Input:
    ------
    graph_template : nested object
    dataframe : pandas data frame

    Output:
    -------
    layouts_of_graphs : nested object
    """

    layouts_of_graphs = []
    curve_dyn_var = None

    for curve_template in graph_template["curves"]:
        if "*" in curve_template["var"]:
            curve_dyn_var = curve_template["var"]
            break

    if not curve_dyn_var:
        graph_layout = deepcopy(graph_template)
        layouts_of_graphs.append(graph_layout)
    else:
        for data_name in dataframe.keys():
            curve_dyn_var_without_star = curve_dyn_var.replace("*", "")
            data_name_spec = string_minus_string(curve_dyn_var_without_star,
                                                 data_name)
            graph_layout = replace_stars_in_graph_template(graph_template,
                                                           data_name_spec,
                                                           dataframe)
            if graph_layout:
                layouts_of_graphs.append(graph_layout)

    return layouts_of_graphs


def replace_stars_in_graph_template(graph_template, data_name_spec, dataframe):
    """It tries to replace the stars in the provided graph template
    by the provided data name specification and thanks to the
    provided data frame

    If it fails, it returns "None"
    Otherwise, it returns the graph layout resulting of the provided
    graph template and the provided data name specification

    Input:
    ------
    graph_template : nested object
    data_name_spec : list of strings
    dataframe : pandas data frame

    Output:
    -------
    graph_layout : nested object
    """

    graph_template_cp = deepcopy(graph_template)
    graph_layout = None

    for curve_template in graph_template_cp["curves"]:
        success = replace_stars_in_curve_template(curve_template,
                                                  data_name_spec,
                                                  dataframe)
        if not success:
            break
        if curve_template is graph_template_cp["curves"][-1]:
            if (graph_template != graph_template_cp and
                    "title" in graph_template_cp and
                    "*" in graph_template_cp["title"] and
                    graph_template_cp["title"].count("*") == len(data_name_spec)):
                graph_template_cp["title"] = replace_stars_in_string(graph_template_cp["title"],
                                                                     data_name_spec)
            graph_layout = graph_template_cp

    return graph_layout


def replace_stars_in_curve_template(curve_template, data_name_spec, dataframe):
    """It tries to replace the stars in the provided curve template
    by the provided data name specification and thanks to the
    provided data frame

    If it fails, it returns "False"
    Otherwise, it returns "True"

    Input:
    ------
    curve_template : dictionary
    data_name_spec : list of strings
    dataframe : pandas data frame

    Output:
    -------
    success : boolean
    """

    success = True

    if "*" in curve_template["var"]:
        if curve_template["var"].count("*") == len(data_name_spec):
            curve_template["var"] = replace_stars_in_string(curve_template["var"],
                                                            data_name_spec)
            if curve_template["var"] in dataframe.keys():
                if ("legend" in curve_template and
                        "*" in curve_template["legend"]
                        and curve_template["legend"].count("*") == len(data_name_spec)):
                    curve_template["legend"] = replace_stars_in_string(curve_template["legend"],
                                                                       data_name_spec)
            else:
                success = False
        else:
            success = False

    return success



"""
shell module implemented as parent for CartShell and AxiShell classes
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from scipy import interpolate, spatial
import matplotlib.pyplot as plt
from arnica.utils import nparray2xmf

MAX_SPLINE_ORDER = 3
SPLINE_SMOOTHNESS = 0

AXIS_NAMES = ['time', 'v', 'u']

class Shell():
    #pylint: disable=too-many-instance-attributes
    """
    *Parent class for computationnal shells*

    :param n_axe_1: Number of x/longitudinal shell points
    :type n_axe_1: int
    :param n_axe_1: Number of y/azimuthal shell points
    :type n_axe_1: int

    Public attributes :

        - **shape** - Shape of the shell : (n_axe_1, n_axe_2)
        - **width_matrix** - Dict containing thickness matrices of shape 'shape'
        - **matrix** - Dict containing fields of shape 'shape' - To deprecate

    Private attributes :

        - **_xyz**
        - **_rad**
        - **_theta**
        - **_n_x**
        - **_n_r**
        - **_n_y**
        - **_n_z**
        - **_du**
        - **_dv**
        - **_abs_curv**
        - **_dwu**
        - **_dwv**
        - **_surf**

    """

    def __init__(self, n_axe_1, n_axe_2):
        """
        *Initialize a Shell object*
        """

        self.shape = (n_axe_1, n_axe_2)
        self.axis_names = AXIS_NAMES

        self._xyz = np.empty((*self.shape, 3))
        self._rad = np.empty(self.shape)
        self._theta = np.empty(self.shape)
        self._n_x = np.empty(self.shape)
        self._n_r = np.empty(self.shape)
        self._n_y = np.empty(self.shape)
        self._n_z = np.empty(self.shape)
        self._du = np.empty(self.shape)
        self._dv = np.empty(self.shape)
        self._abs_curv = np.empty(self.shape[0])
        self._dwu = np.empty(self.shape)
        self._dwv = np.empty(self.shape)
        self._surf = np.empty(self.shape)

        u_range = np.linspace(0., 1., num=self.shape[0])
        v_range = np.linspace(0., 1., num=self.shape[1])
        self._u_adim = np.ones(self.shape) * u_range[:, np.newaxis]
        self._v_adim = np.ones(self.shape) * v_range[np.newaxis,:]

        self.width_matrix = {}

    @property
    def xyz(self):
        """
        *Getter method for xyz matrix of shape (self.shape, 3)*
        """
        return self._xyz

    @property
    def rad(self):
        """
        *Getter method for rad matrix of shape (self.shape,)*
        """
        return self._rad

    @property
    def theta(self):
        """
        *Getter method for theta matrix of shape (self.shape,)*
        """
        return self._theta

    @property
    def n_x(self):
        """
        *Getter method for x-normal matrix of shape (self.shape,)*
        """
        return self._n_x

    @property
    def n_r(self):
        """
        *Getter method for r-normal matrix of shape (self.shape,)*
        """
        return self._n_r

    @property
    def n_y(self):
        """
        *Getter method for y-normal matrix of shape (self.shape,)*
        """
        return self._n_y

    @property
    def n_z(self):
        """
        *Getter method for z-normal matrix of shape (self.shape,)*
        """
        return self._n_z

    @property
    def du(self):
        #pylint: disable=invalid-name
        """
        *Getter method for u-direction spacing matrix of shape (self.shape,)*
        """
        return self._du

    @property
    def dv(self):
        #pylint: disable=invalid-name
        """
        *Getter method for v-direction spacing matrix of shape (self.shape,)*
        """
        return self._dv

    @property
    def u_adim(self):
        #pylint: disable=invalid-name
        """
        *Getter method for u-direction non dimentional matrix of shape (self.shape,)*
        """
        return self._u_adim

    @property
    def v_adim(self):
        #pylint: disable=invalid-name
        """
        *Getter method for v-direction non dimentional matrix of shape (self.shape,)*
        """
        return self._v_adim

    @property
    def abs_curv(self):
        """
        *Getter method for central curvilinear abscissa array of shape (self.shape[1],)*
        """
        return self._abs_curv

    @property
    def dwu(self):
        #pylint: disable=invalid-name
        """
        *Getter method for weighted u-direction spacing matrix of shape (self.shape,)*
        """
        return self._dwu

    @property
    def dwv(self):
        #pylint: disable=invalid-name
        """
        *Getter method for weighted v-direction spacing matrix of shape (self.shape,)*
        """
        return self._dwv

    @property
    def surf(self):
        #pylint: disable=invalid-name
        """
        *Getter method for weighted surface matrix of shape (self.shape,)*
        """
        return self._surf

    def dump_shell(self, name='shell'):
        """
        *Dump shell geometrical properties into hdf file*

        Optional :
        :param name: Name of the file to dump. Default 'shell'
        :type name: str
        """
        xmdf = nparray2xmf.NpArray2Xmf("./" + name + ".h5")
        xmdf.create_grid(np.take(self._xyz, 0, axis=-1),
                         np.take(self._xyz, 1, axis=-1),
                         np.take(self._xyz, 2, axis=-1))
        xmdf.add_field(self._rad, 'r')
        xmdf.add_field(self._theta, 'theta')
        xmdf.add_field(self._n_x, 'n_x')
        xmdf.add_field(self._n_y, 'n_y')
        xmdf.add_field(self._n_z, 'n_z')
        xmdf.add_field(self._n_r, 'n_r')
        xmdf.add_field(self._du, 'du')
        xmdf.add_field(self._dv, 'dv')
        xmdf.add_field(self._u_adim, 'u_adim')
        xmdf.add_field(self._v_adim, 'v_adim')
        xmdf.add_field(self._dwu, 'dwu')
        xmdf.add_field(self._dwv, 'dwv')
        xmdf.add_field(self._surf, 'surf')

        xmdf.dump()

    def add_curviwidth(self, label, points):
        """
        *Add a 2D width matrix of shell shape extruded from points spline*

        The witdh matrix is iso-shape[0]

        :param label: Label of the width matrix
        :type label: str
        :param points: Tuple (dim n) of tuple (dim 2) of float coordinates\

        Optional :
        :param replace: If True, overwrite matrix.
                        If False, addition matrix to possible previous one.
                        Default True
        :type replace: bool
        """

        (x_tuple, y_tuple) = tuple(zip(*points))

        # Extend Bounds
        xlist = [0] + list(x_tuple) + [1]
        ylist = [y_tuple[0]] + list(y_tuple) + [y_tuple[-1]]
        # Generate continuous fictive spline
        f_int = interpolate.interp1d(xlist, ylist)
        # Generate discretise spline
        xnew = np.linspace(0, 1, num=self.shape[1])
        ynew = f_int(xnew)

        if label not in self.width_matrix:
            self.width_matrix[label] = np.zeros(self.shape)

        # Extrude azimuthally
        self.width_matrix[label] += np.tile(ynew, (self.shape[0], 1))

    def set_mask_on_shell(self, point_cloud, tol):
        """
        *Create a mask on the shell from a point cloud*

        The mask value is 0 for shell points located near cloud points.
        Otherwise the mask value is 1.

        :param point_cloud: Array of dim (n,3) of coordinates of points.
        :type point_cloud: numpy array
        :param tol: Tolerance of proximity
        :type tol: int
        """

        # Create a KDTree from cloud points
        kdtree = spatial.KDTree(point_cloud)
        # Compute distances between cloud points and
        # the closest neighbor from the 2D shell points.
        dists, _ = kdtree.query(self.xyz, k=1)
        # Mask takes 1 if the distance is below tolerance.
        mask = np.where(dists.reshape(self.shape) > tol,
                        1.,
                        0.)

        return mask

    def bake_millefeuille(self, width_matrix_label, n_layers, shift=0.0):
        """
        *Create a millefeuille-like shell.

        Extrude a 2D shell in the normal direction up
        pointwise height given by "width_matrix_label" matrix.

        :param width_matrix_label: Label of the width matrix
        :type width_matrix_label: str
        :param n_layers: Number of layer for extrusion
        :type n_layers: int
        :param shift: Additional depth (optional)
        :type shift: float

        :returns:

            - **cake** - A dict() containing shell data :

                - *xyz* - np.array of dim (self.shape, n_layers, 3)
                - *dz*  - np.array of dim (self.shape, n_layers)

        "Bon appetit!"
        """

        cake = {}
        cake["xyz"] = np.empty((self.shape[0],
                                self.shape[1],
                                n_layers,
                                3))
        cake["dz"] = np.empty((self.shape[0],
                               self.shape[1],
                               n_layers))

        for j, n_j in enumerate((self._n_x, self._n_y, self._n_z)):
            for i in range(n_layers):
                cake["xyz"][:, :, i, j] = (np.take(self._xyz, j, -1)
                                           + (1.0 * i / n_layers
                                              * self.width_matrix[width_matrix_label]
                                              + shift)
                                           * n_j)

        for i in range(n_layers):
            cake["dz"][:, :, i] = abs(self.width_matrix[width_matrix_label] / n_layers)
        cake["dz"][:, :, 0] /= 2
        cake["dz"][:, :, -1] /= 2

        return cake

    def average_on_shell_over_dirs(self, variable, directions, scale=True):
        """
        *Performs an integration (average) over one or multiple directions*

        :param variable: A np.array to be averaged of dim (n_time, n_v, n_u)\
        :param directions: A list() of directions on which the average process\
                           is to be performed.\
                           Contains keywords from ['time','v','u'].

        :returns:

            - **averaged_variable** - A np.array of averaged data on given directions.
            - **min_variable** - A np.array of min data on given directions.
            - **max_variable** - A np.array of max data on given directions.
        """

        dirs = self.axis_names.copy()
        if scale:
            averaged_variable = np.multiply(variable, self.surf)
        else:
            averaged_variable = variable

        if (len(directions) > 0 and
                all(dir_ in dirs for dir_ in directions)):
            for dir_ in directions:
                index = dirs.index(dir_)
                averaged_variable = np.mean(averaged_variable, axis=index)
                dirs.pop(index)

        else:
            mess = 'Averaging directions do not conform to criteria\n'
            mess += 'It should be a list containing one of or all items\n'
            for axis in self.axis_names:
                mess += axis + ', '
            raise ValueError(mess)

        return averaged_variable

    def plot(self, savefile=None):
        """
        *Plot 2D curve of the Shell*

        Optional :
        :param savefile: If None, the figure is plot.
                         If string, the figure is saved into files
        """

        plt.plot(self.xyz[0, :, 0], self.xyz[0, :, 1])
        plt.title("Shell")
        plt.xlabel("x")
        plt.ylabel("y")

        if savefile is None:
            plt.show()
        else:
            plt.savefig(savefile)

    def _compute_shell_crest(self, ctrl_pts_1, ctrl_pts_2):
        """
        *Build the shell crest of shape (2, self.shape[1]) and respecting controle points*
        """

        # Build crest
        spline_order = min(len(ctrl_pts_1) - 1, MAX_SPLINE_ORDER)
        # Generate continuous spline from control points.
        tck, _ = interpolate.splprep([ctrl_pts_1,
                                      ctrl_pts_2],
                                     s=SPLINE_SMOOTHNESS,
                                     k=spline_order)
        unew = np.linspace(0, 1, num=self.shape[1])
        # Generate discrete spline
        # Numpy Array of dim (2, self.shape[1]) [x,r] / [x,y]
        shell_crest = np.asarray(interpolate.splev(unew, tck))

        return shell_crest

    def _compute_shellcrest_nml(self, shell_crest):
        #pylint: disable=no-self-use
        """
        *Private methold returning shell crest normals*

            - AxiShell -> x,r-normal
            - CartShell -> x,y-normal
        """
        nml_vect = np.roll(np.diff(shell_crest), 1, axis=0)
        nml_vect /= np.linalg.norm(nml_vect, axis=0)
        nml_vect[0, :] *= -1
        nml_vect = np.pad(nml_vect, ((0, 0), (0, 1)), mode='edge')
        return nml_vect



"""
string_manager.py

Functions which deal with strings
"""

from difflib import ndiff

__all__ = []


def split_string_around_stars(string):
    """It splits the provided string around every star in it
    and returns the result in the form of a list

    Input:
    ------
    string : string

    Output:
    -------
    splitted_string : list
    """

    splitted_string = []
    word = ""

    for char in string:
        if char == '*':
            if word:
                splitted_string.append(word)
                word = ""
            splitted_string.append('*')
        else:
            word += char

    if word:
        splitted_string.append(word)

    return splitted_string


def string_minus_string(string1, string2):
    """It returns the strings we have to add to "string1" to get it directly
    equal to "string2" in cases in which "string2" contains "string1" by parts
    plus some differences

    Input:
    ------
    string1 : string
    string2 : string

    Output:
    -------
    differences : list of strings
    """

    differences = []
    chars_differences = list(ndiff(string1, string2))
    chars_differences = list(enumerate([[position, char_diff[-1]]
                                        for position, char_diff in
                                        enumerate(chars_differences)
                                        if char_diff[0] == "+"]))

    if chars_differences:
        differences.append(chars_differences[0][1][1])
        for ttuple in chars_differences[1:]:
            index = ttuple[0]
            position = ttuple[1][0]
            character = ttuple[1][1]
            last_position = chars_differences[index-1][1][0]
            if position == last_position+1:
                differences[-1] += character
            else:
                differences.append(character)

    return differences


def replace_stars_in_string(string, str_list):
    """It replaces every star in "string" by its associated string
    in "str_list" and returns the new string created

    Input:
    ------
    string : string
    str_list : list of strings

    Output:
    -------
    new_string : string
    """

    splitted_string = split_string_around_stars(string)
    index = 0

    for position, sstring in enumerate(splitted_string):
        if sstring == '*':
            splitted_string[position] = str_list[index]
            index += 1

    new_string = "".join(splitted_string)

    return new_string



""" axishell module
to create x-axisymmetric shells for scientific computations
"""

import numpy as np
from arnica.utils.shell import Shell

class CartShell(Shell):
    #pylint: disable=too-many-instance-attributes
    r"""\
    *Base class for cartesian computationnal shells*

    CartShell class is based on Shell class, inheriting of Shell methods.
    At initialization, the shell is builded an an cartesian shell.

    Some attributes are based on **u** and **v**, defined as the curvilinear
    longitudinal (u) abscissa and curvilinear transversal (v) abscissa
    respectively.

    The shell is set using three corners (assuming they are othogonal)

    ::
          X umax
          ________________
          |              |            Cartesian system
          |              |            Example longi/u <=> x  and v <==> z
          |              |            zero. (0, 0, 0)
          |              |            umax. (1, 0, 0)
          |              |            vmax. (0, 0, 1)
          |              |            x - longi/u
          |              |            ^
          |              |            |
          |              |            |
          |              |            O-----> z - transvers/v
          |______________| X vmax     y
          X zero.

    :param n_transvers: Number of azimuthal shell points
    :type n_transvers: int
    :param n_longi: Number of longitudinal shell points
    :type n_longi: int
    :param z_range: Extrusion range in transvers direction
    :type z_range: float
    :param zero: Array of dim (3) lower left corner of shell
    :param umax: Array of dim (3) corner of shell for umax
    :param vmax: Array of dim (3) corner of shell for vmax

    Public attributes :

        - **shape** - Shape of the shell : (n_transvers, n_longi)
        - **width_matrix** - Dict containing thickness matrices of shape 'shape'
        - **matrix** - Dict containing fields of shape 'shape' - To deprecate

    Private attributes :

        - **_xyz**
        - **_rad**
        - **_theta**
        - **_n_x**
        - **_n_r**
        - **_n_y**
        - **_n_z**
        - **_du**
        - **_dv**
        - **_abs_curv**
        - **_dwu**
        - **_dwv**
        - **_surf**

    """

    def __init__(self,
                 n_trans,
                 n_longi,
                 zero,
                 umax,
                 vmax,
                 ):
        #pylint: disable=too-many-arguments
        """
        *Initialize a CartShell object*
        """

        Shell.__init__(self, n_trans, n_longi)
        self._build_shell(
            np.array(zero),
            np.array(umax),
            np.array(vmax))

    def _build_shell(self, zero, umax, vmax):
        #pylint: disable=arguments-differ
        """
        *Build shell from geometric features*

            - Construct a spline used as base for extrusion\
              from control points : tck
            - Discretise the spline : shell_crest
            - Compute normal vectors for the 1D shell_crest
            - Compute r,n_x,n_r-components for 2D shell
            - Compute theta-components for 2D shell
            - Compute xyz,n_y,n_z-components for 2D shell
        """

        vec_u = umax-zero
        vec_v = vmax-zero

        norm = np.cross(vec_u, vec_v)
        norm /= np.linalg.norm(norm)

        u_range = np.linspace(0., 1., num=self.shape[0])
        v_range = np.linspace(0., 1., num=self.shape[1])
        d_u = np.linalg.norm(vec_u)/(self.shape[0]-1)
        d_v = np.linalg.norm(vec_v)/(self.shape[1]-1)

        v_coor, u_coor = np.meshgrid(v_range, u_range)
        u_coor = vec_u[np.newaxis, np.newaxis, :]*u_coor[:, :, np.newaxis]
        v_coor = vec_v[np.newaxis, np.newaxis, :]*v_coor[:, :, np.newaxis]
        tmp_x = (
            zero[np.newaxis, np.newaxis, 0]
            + u_coor[:, :, 0] + v_coor[:, :, 0])
        tmp_y = (
            zero[np.newaxis, np.newaxis, 1]
            + u_coor[:, :, 1] + v_coor[:, :, 1])
        tmp_z = (
            zero[np.newaxis, np.newaxis, 2]
            + u_coor[:, :, 2] + v_coor[:, :, 2])

        self._xyz = np.stack((tmp_x, tmp_y, tmp_z), axis=-1)

        # Compute radius and azimuth matrix of shape (n_transvers, n_longi)
        self._rad = np.sqrt(
            np.take(self._xyz, 1, -1) ** 2
            + np.take(self._xyz, 2, -1) ** 2)
        self._theta = np.arctan2(
            np.take(self._xyz, 2, -1),
            np.take(self._xyz, 1, -1))

        # Compute x,y,z,r-normal matrix of shape (n_transvers, n_longi)
        self._n_x = np.full(self.shape, norm[0])
        self._n_y = np.full(self.shape, norm[1])
        self._n_z = np.full(self.shape, norm[2])
        self._n_r = self._n_y * np.cos(self._theta)

        self._du = np.full(self.shape, d_u)
        self._dv = np.full(self.shape, d_v)

        # Compute abs_curv array of shape (n_longi,)
        self._abs_curv = np.cumsum(np.take(self._du, 0, 0)) - self._du[0, 0]

        # Compute weight intervals in u and v directions array of shape (n_transvers, n_longi)
        self._dwu = self._du.copy()
        self._dwu[:, (0, -1)] /= 2
        self._dwv = self._dv.copy()
        self._dwv[(0, -1), :] /= 2

        # Compute surface nodoes array of shape (n_transvers, n_longi)
        self._surf = np.multiply(self._dwu, self._dwv)




"""interpolate a cloud from an other cloud """

import numpy as np
from scipy import spatial
# from arnica.utils.vector_actions import (yz_to_theta)

SMALL = 1e-16

def _build_ravel_component(in_xyz, limitsize=None):
    """ build source and target raveled components

    Parameters:
    -----------
    in_xyz (dict of nparray): las index are coordinantes 3D. (:,:,3)
    limitsize : maximum size of raveled output

    Returns:
    --------
    out_xyz : raveld version, limited in size
    """
    skipper = 1
    if limitsize is not None:
        input_size = np.ravel(np.take(in_xyz, 0, axis=-1)).shape[0]
        skipper = max(int(input_size * 1.0 / limitsize), 1)
        if skipper > 1:
            print("Warning, source data too big ({0}/{1}),\n "
                  "sub sampling every {2}th element.".format(input_size,
                                                             limitsize,
                                                             skipper))
    out_xyz = np.stack((np.ravel(np.take(in_xyz, 0, axis=-1))[::skipper],
                        np.ravel(np.take(in_xyz, 1, axis=-1))[::skipper],
                        np.ravel(np.take(in_xyz, 2, axis=-1))[::skipper]),
                       axis=1
                      )

    return out_xyz, skipper


def _interpolate_var(in_val, index, inv_dist):
    """interpolate nP_array of source
    intour a np.arry of target"""

    estimate = (np.sum(inv_dist *
                       np.reshape(in_val[np.ravel(index)],
                                  np.shape(inv_dist)),
                       axis=1))

    estimate /= np.sum(inv_dist, axis=1)

    return estimate

def _compute_dists(source_xyz, target_xyz, limitsource, stencil, power):
    sce_ravel, skipper = _build_ravel_component(source_xyz,
                                                limitsize=limitsource)
    tgt_ravel, _ = _build_ravel_component(target_xyz)

    kdtree = spatial.cKDTree(sce_ravel) #pylint: disable=not-callable

    dists, index = kdtree.query(tgt_ravel, k=stencil,)

    if power != 1.0:
        dists = np.power(dists, power)

    inv_dist = np.reciprocal(np.maximum(dists, SMALL))
    return index, dists, inv_dist, skipper

def cloud2cloud(source_xyz, source_val, target_xyz, stencil=3, **kwargs):
    """ Interpolate  form a cloud to an other

    Parameters :
    ------------
    source_xyz : numpy array shape (n_s, 3) either (1000, 3 )  or (10,10,10, 3)
    source_val : numpy array shape (n_s, k) of k variables
    target_xyz : numpy array shape (n_t, 3)
    stencil (int): nb of neigbors to compute (1 is closest point)

    Optional keyword arguments
    --------------------------
    limitsource (int) : maximum nb of source points allowed (subsample beyond)
    power(float) : Description
    tol(float) : Description
    Returns :
    ----------
    target_val : numpy array shape (n_t, k)

    """

    opts = {'limitsource':None, 'power':1.0, 'tol':None}
    for keyword in opts:
        if keyword in kwargs:
            opts[keyword] = kwargs[keyword]

    index, dists, inv_dist, skipper = _compute_dists(source_xyz, target_xyz,
                                                     opts['limitsource'],
                                                     stencil,
                                                     opts['power'])
    target_val = {}
    for key in source_val:
        if stencil > 1:
            estimate = _interpolate_var(np.ravel(source_val[key])[::skipper],
                                        index,
                                        inv_dist,
                                        )
        else:
            estimate = np.ravel(source_val[key])[::skipper][index]

        if opts['tol'] is not None:
            estimate = np.where(dists[:, 0] > opts['tol'], 0, estimate)

        target_val[key] = np.reshape(estimate,
                                     np.take(target_xyz, 0, axis=-1).shape)


    return target_val



"""
showy.py

Showy
=====


**SHOWY**  in `arnica/utils` is a helper for matplotlib subplots.
If the data is stored as a disct, the layout can be saved as a template
in .yml format.
If can use wildcards if the dictionary keys allows it.


Simple example
***************


::

    import numpy as np
    from arnica.utils.matplotlib_display import showy


    def showy_demo_plain():
        data = dict()
        data["time"] = np.linspace(0, 0.1, num=256)

        data["sine_10"] = np.cos(data["time"] * 10 * 2 * np.pi)
        data["sine_30"] = np.cos(data["time"] * 30 * 2 * np.pi)
        data["sine_100"] = np.cos(data["time"] * 100 * 2 * np.pi)
        data["sine_100p1"] = 1. + np.cos(data["time"] * 100 * 2 * np.pi)

        # Creating a template
        layout = {
            "title": "Example",
            "graphs": [
                {
                    "curves": [{"var": "sine_10"}],
                    "x_var": "time",
                    "y_label": "Fifi [mol/m³/s]",
                    "x_label": "Time [s]",
                    "title": "Sinus of frquency *"
                },
                {
                    "curves": [{"var": "sine_30"}],
                    "x_var": "time",
                    "y_label": "Riri [Hz]",
                    "x_label": "Time [s]",
                    "title": "Second graph"
                },
                {
                    "curves": [
                        {
                            "var": "sine_100",
                            "legend": "origin",
                        },
                        {
                            "var": "sine_100p1",
                            "legend": "shifted",
                        }
                    ],
                    "x_var": "time",
                    "y_label": "Loulou [cow/mug]",
                    "x_label": "Time [s]",
                    "title": "Third graphg"
                }
            ],
            "figure_structure": [3, 1],
            "figure_dpi": 92.6
        }

    # Displaying the data described in the new created layout
    showy(layout, data)


Using wildcard ' * '
********************

In showy you can show all the graphs with a same prefix putting a "**\***".
For example if you have 3 variables like var_1, var_2, var_3
you can just write var_*. An example is shown below:

::

    import numpy as np
    from arnica.utils.matplotlib_display import display

    def showy_demo_wildcards():
        data = dict()
        data["time"] = np.linspace(0, 0.1, num=256)

        freq = 10.
        for freq in np.linspace(10, 20, num=9):
            data["sine_" + str(freq)] = np.cos(data["time"]*freq*2*np.pi)

        # Creating a template
        template = {
            "title": "Example",
            "graphs": [{
                "curves": [{"var": "sine_*"}],
                "x_var": "time",
                "y_label": "Sine [mol/m³/s]",
                "x_label": "Time [s]",
                "title": "Sinus of frquency *"
            }],
            "figure_structure": [3, 3],
            "figure_dpi": 92.6
        }

        showy(template, data)


Options available
***************

The scheme that showing all the options available is shown below.

.. highlight:: yaml

::

    title: Layout scheme
    description: The structure that a layout has to respect in order to be used to plot
      data with Showy
    type: object
    properties:
      title:
        description: The title of the layout
        type: string
      graphs:
        description: The graphs of the layout
        type: array
        items:
          description: A graph
          type: object
          properties:
            curves:
              description: The curves of the graph
              type: array
              items:
                description: A curve
                type: object
                properties:
                  var:
                    description: The name of the data for the Y-axis
                    type: string
                  legend:
                    description: The legend of the curve
                    type: string
                required:
                - var
                additionalProperties: false
              minItems: 1
            x_var:
              description: The name of the data for the X-axis
              type: string
            y_label:
              description: The label of the Y-axis
              type: string
            x_label:
              description: The label of the X-axis
              type: string
            title:
              description: The title of the graph
              type: string
          required:
          - curves
          - x_var
          additionalProperties: false
        minItems: 1
      figure_dpi:
        description: The number of dots per inch of the figure
        type: number
        exclusiveMinimum: 0
      figure_size:
        description: The size of the figure in inches
        type: array
        items:
          description: A length in inches
          type: number
          exclusiveMinimum: 0
        minItems: 2
        maxItems: 2
      figure_structure:
        description: The numbers of rows and columns of graphs
        type: array
        items:
          description: An integer for a number of rows or columns
          type: integer
          minimum: 1
        minItems: 2
        maxItems: 2
    required:
    - graphs
    additionalProperties: false

"""

import warnings
from math import ceil

import pandas as pd
import matplotlib.pyplot as plt
from arnica.utils.lay_and_temp_manager import decompact_template

try:
    from mplcursors import cursor
except:
    pass

__all__ = ["showy", "display"]


def display(**kwargs):
    """Retro compatibility"""

    warnings.warn("Use *showy* instead of *display*", DeprecationWarning)
    showy(**kwargs)


def showy(layout, data, show=True):
    """It displays the desired graphs described by the provided layout
    thanks to the provided key-value object which contains
    the required data

    Input:
    ------
    data : key-value object
    layout : nested object
    """

    for curve in layout["graphs"]:
        for var in curve["curves"]:
            for val in var.values():
                if "*" in val:
                    layout = decompact_template(layout, data)

    if not isinstance(data, pd.DataFrame):
        dataframe = pd.DataFrame.from_dict(data)
    else:
        dataframe = data

    figure_number = 0
    graph_position = 1
    display_params = initialize_display(layout)

    for graph_layout in layout['graphs']:
        if (graph_position > display_params['numb_max_of_graphs_per_fig']
                or figure_number == 0):
            graph_position = 1
            figure_number += 1
            figure = build_new_figure(display_params, figure_number)
        build_new_graph(figure, display_params['fig_struct'],
                        graph_position, graph_layout, dataframe)
        graph_position += 1

    try:
        cursor()
    except:
        pass

    if show:
        plt.show()


def initialize_display(layout):
    """Sets some parameters needed by the function "display", thanks
    to the provided layout, and returns them in a dictionary

    Input:
    ------
    layout : nested object

    Output:
    ------
    display_params : dictionary
    """

    display_params = {}
    display_params['numb_of_graphs'] = len(layout['graphs'])

    if "figure_structure" in layout:
        display_params['fig_struct'] = layout["figure_structure"]
    else:
        if display_params['numb_of_graphs'] == 1:
            display_params['fig_struct'] = [1, 1]
        else:
            display_params['fig_struct'] = [2, 3]

    display_params['numb_max_of_graphs_per_fig'] = (
        display_params['fig_struct'][0] * display_params['fig_struct'][1])
    display_params['numb_of_figs'] = ceil(
        display_params['numb_of_graphs']
        / display_params['numb_max_of_graphs_per_fig'])

    if "figure_dpi" in layout:
        display_params['fig_dpi'] = layout["figure_dpi"]
    else:
        display_params['fig_dpi'] = None

    if "figure_size" in layout:
        display_params['fig_size'] = layout["figure_size"]
    else:
        display_params['fig_size'] = None

    if "title" in layout:
        display_params['lay_title'] = layout['title']
    else:
        display_params['lay_title'] = None

    return display_params


def build_new_figure(display_params, figure_number):
    """Builds a new matplotlib figure thanks to the provided parameters

    Input:
    ------
    display_params : dictionary
    figure_number : integer

    Output:
    ------
    figure : matplotlib object
    """

    if display_params['fig_dpi']:
        figure = plt.figure(dpi=display_params['fig_dpi'])
    else:
        figure = plt.figure()

    if display_params['fig_size']:
        figure.set_size_inches(*display_params['fig_size'])

    if display_params['lay_title']:
        if display_params['numb_of_graphs'] <= (
                display_params['numb_max_of_graphs_per_fig']):
            figure_name = display_params['lay_title']
        else:
            figure_name = (display_params['lay_title'] +
                           ' (' + str(figure_number) + '/' +
                           str(display_params['numb_of_figs']) + ')')
        figure.canvas.set_window_title(figure_name)
        figure.suptitle(figure_name, fontsize=20,
                        fontweight='bold', y=0.99)

    figure.subplots_adjust(left=0.1, bottom=0.1,
                           right=0.9, top=0.9,
                           wspace=0.3, hspace=0.3)

    return figure


def build_new_graph(
        figure,
        figure_structure,
        graph_position,
        graph_layout,
        dataframe):
    """Builds a new matplotlib graph in the provided matplotlib
    figure thanks to the provided parameters

    Input:
    ------
    figure : matplotlib object
    figure_structure : tuple of 2 integers
    graph_position : integer
    graph_layout : nested object
    dataframe : pandas data frame
    """

    subplot = figure.add_subplot(*figure_structure, graph_position)

    if "x_label" in graph_layout:
        subplot.set_xlabel(graph_layout["x_label"])

    if "y_label" in graph_layout:
        subplot.set_ylabel(graph_layout["y_label"])

    if "title" in graph_layout:
        subplot.set_title(graph_layout["title"],
                          fontweight='bold')

    if graph_layout["x_var"] in dataframe.keys():
        for curve_layout in graph_layout["curves"]:
            if curve_layout["var"] in dataframe.keys():
                if "legend" in curve_layout:
                    curve_legend = curve_layout["legend"]
                else:
                    curve_legend = curve_layout["var"]
                subplot.plot(dataframe[graph_layout["x_var"]],
                             dataframe[curve_layout["var"]],
                             label=curve_legend)

    subplot.legend(loc='upper right')



import numpy as np
import matplotlib
from matplotlib.ticker import FuncFormatter
import matplotlib.pyplot as plt
import scipy.stats as stats


def resample_signal(time, signal, dtime=None):
    """
	*Resample the initial signal at a constant time interval.*
  
	:param time: Time vector of your signal
	:param signal: Signal vector
	:dtime: New time step
	:returns:            
	- **rescaled_time** - Uniformally rescaled time vector
	- **rescaled_signal** - Rescaled signal

	.. note:: 	If a dtime is given, the interpolation is made to 
				have a signal with a time interval of dtime.
				Else, the dt is the smallest time interval between
				two values of the signal.
	"""

    if dtime is None:
        time_steps = np.diff(time)
        dtime = time_steps.min()

    rescaled_time = time - time[0]
    rescaled_time = np.arange(rescaled_time[0], rescaled_time[-1], dtime)

    rescaled_signal = np.interp(rescaled_time, time - time[0], signal)

    return rescaled_time, rescaled_signal


def calc_autocorrelation_time(time, signal, threshold=0.2):
    """
	*Estimate the autocorrelation time at a given threshold.*
	
	:param time: Time vector of your signal
	:param signal: Signal vector
	:param threshold: Threshold under which the signal is correlated
	:type threshold: float

	:returns:

	- **autocorrelation_time** - Minimum time step to capture the signal at a
								 correltion under thethreshold
	"""

    signal -= np.mean(signal)
    correlation = np.correlate(signal, signal, "full")
    correlation = abs(correlation[int((correlation.size - 1) / 2) :])
    correlation /= correlation[0]

    autocorrelation_time = time[np.argmax(correlation < threshold)]

    return autocorrelation_time


def show_autocorrelation_time(time, signal, threshold=0.2):
    """
	*Plot the autocorrelation function of the signal.*
	
	:param time: Time vector of your signal
	:param signal: Signal vector
	:param threshold: Autocorrelation threshold

	:returns:

	- **fig** - Figure of the result

	"""

    signal -= np.mean(signal)
    correlation = np.correlate(signal, signal, "full")
    correlation = abs(correlation[int((correlation.size - 1) / 2) :])
    correlation /= correlation[0]

    autocorrelation_time = time[np.argmax(correlation < threshold)]

    figout = plt.figure()
    plt.plot(time, correlation)
    plt.axvline(autocorrelation_time, c="r")
    plt.xlabel("Time (s)")
    plt.ylabel("Correlation")
    plt.semilogx()
    plt.grid()

    return figout


def sort_spectral_power(time, signal):
    """
	*Determine the harmonic power contribution of the signal.*
	
	:param time: Time vector of your signal
	:param signal: Signal vector

	returns:

	- **harmonic_power** - Harmonic power of the signal
	- **total_power** - Total spectral power of the signal


	.. note:: 	It calculates the Power Spectral Density (PSD) of the complete
				signal and of a downsampled version of the signal. The 
				difference of the two PSD contains only harmonic components.
	"""

    frequency1, psd1 = power_spectral_density(time, signal)
    deltaf1 = 1.0 / ((time[1] - time[0]) * 2)

    downsampled_time = time[0:-1:2]
    downsampled_signal = signal[0:-1:2]
    frequency2, psd2 = power_spectral_density(downsampled_time, downsampled_signal)
    deltaf2 = 1.0 / ((downsampled_time[1] - downsampled_time[0]) * 2)

    psd2 = np.interp(frequency1, frequency2, psd2) * 2

    harmonic_power = np.cumsum(np.abs(psd1 - psd2) * 2)[-1] * deltaf1
    total_power = np.cumsum(psd1)[-1] * deltaf1

    return harmonic_power, total_power


def power_representative_frequency(time, signal, threshold=0.8):
    """
	*Calculate the frequency that captures a level of spectral power.*

	:param time: Time vector of your signal
	:param signal: Signal vector
	:param threshold: Level of representativity of the spectral power

	:returns:

	- **representative_frequency** - Frequency above which the power threshold is reached

	.. note:: 	It calculates the cumulative power spectral density and returns
				the frequency that reaches the threshold of spectral power.
	"""

    frequency, power_density = power_spectral_density(time, signal)

    cumulative_psd = np.cumsum(power_density)
    cumulative_psd /= cumulative_psd[-1]
    representative_frequency = frequency[np.argmax(cumulative_psd > threshold)]

    return representative_frequency


def show_power_representative_frequency(time, signal, threshold=0.80):
    """
	*Plot the power spectral density of the signal.*
	
	:param time: Time vector of your signal
	:param signal: Signal vector
	:param threshold: Power representative frequency threshold

	:returns:

	- **fig** - Figure of the result

	"""

    frequency, power_density = power_spectral_density(time, signal)
    cumulative_psd = np.cumsum(power_density)
    cumulative_psd /= cumulative_psd[-1]
    representative_frequency = frequency[np.argmax(cumulative_psd > threshold)]

    figout = plt.figure()
    plt.plot(frequency, cumulative_psd)
    plt.axvline(representative_frequency, c="r")
    plt.xlabel("frequency (Hz)")
    plt.ylabel("PSD (%)")
    plt.grid()

    return figout


def ks_test_distrib(data, distribution="normal"):
    """
	*Calculate the correlation score of the signal with the distribution*

	:param data: array of values
	:param distribution: kind of distribution the values follow to test

	:returns:

	- **score** -Minimum score over the height of the ks test
	- **position** -Index of the height at which the min. of the test is found
	- **height** -Corresponding heiht where the min. is found
	- **scale** -Scale parameter of the lognormal fitting

	"""
    data_shape = np.shape(data)
    if len(data_shape) > 1:
        p_value = np.zeros(data_shape[1])
        for j in range(data_shape[1]):
            data_1d = data[:, j]

            mu = np.average(data_1d)
            sigma = np.std(data_1d)

            # if data_1d.min() <= 0:
            # 	p_value[j]=1
            # else:
            if distribution == "lognormal":
                sigma, loc, scale = stats.lognorm.fit(data_1d, floc=0)
                stat, p_value[j] = stats.kstest(
                    data_1d, "lognorm", args=(sigma, loc, scale)
                )
                # if p_value[j]<5.0e-2:
                # 	print(p_value[j])
                # 	print(temperature.min())
                # 	print(temperature.max())

            elif distribution == "normal":
                loc, sigma = stats.norm.fit(data_1d, loc=300)
                stat, p_value[j] = stats.kstest(data_1d, "norm", args=(loc, sigma))
                scale = None
        position = np.argmin(p_value)
        height = position / data_shape[1]
    else:
        p_value = 0

        data_1d = data
        mu = np.average(data_1d)
        sigma = np.std(data_1d)
        if distribution == "lognormal":
            sigma, loc, scale = stats.lognorm.fit(data_1d, floc=0)
            stat, p_value = stats.kstest(data_1d, "lognorm", args=(sigma, loc, scale))
            # if p_value[j]<5.0e-2:
            # 	print(p_value[j])
            # 	print(temperature.min())
            # 	print(temperature.max())

        elif distribution == "normal":
            loc, sigma = stats.norm.fit(data_1d, loc=300)
            stat, p_value = stats.kstest(data_1d, "norm", args=(loc, sigma))
            scale = None
        position = np.argmin(p_value)
        height = None

    # print(p_value)
    score = np.min(p_value)

    return score, position, height, scale


def show_temperature_distribution(temperature_recording, height, distribution="normal"):
    """
	*Plot the temperature distribution and the fitting curve*

	:param temperature_recording: Temperature as a function of height and time
	:param height: Height in the plan40
	:param distribution: Type of distribution for the fitting method

	:returns:

	- **fig** - Figure of the result
	"""
    hH = np.linspace(0, 1, len(temperature_recording[0, :]))
    index = np.abs(hH - height).argmin()
    data = temperature_recording[:, index]

    abscissa = np.linspace(0, np.amax(data), num=50)

    if distribution == "normal":
        loc, sigma = stats.norm.fit(data, loc=300)
        pdf_fitted = stats.norm.pdf(abscissa, loc, sigma)

    if distribution == "lognormal":
        sigma, loc, scale = stats.lognorm.fit(data, floc=300)
        pdf_fitted = stats.lognorm.pdf(abscissa, sigma, loc, scale)

    figout = plt.figure()
    plt.hist(data, density=True, bins=30, color="lightgray")
    plt.plot(abscissa, pdf_fitted)
    plt.xlabel("T (K)")

    return figout


def duration_for_uncertainty(
    time, signal, target=10, confidence=0.95, distribution="normal"
):
    """
	*Give suggestion of simulation duration of a plan40 calculation.*
	
	:param time: Time vector of your signal
	:param signal: Signal vector
	:param target: Desired amplitude of the confidence interval
	:param confidence: Level of confidence of the interval
	:param distribution: Type of distribution of the signal to make the interval

	:returns:

	- **duration** -Duration of the signal
	"""

    dtime = calc_autocorrelation_time(time, signal)
    sigma = np.std(signal)

    if distribution == "normal":
        z_coef = stats.norm.ppf(1 - (1 - confidence) / 2)
        n_timesteps = int((2 * z_coef * sigma / target) ** 2)

    duration = dtime * n_timesteps

    return duration


def uncertainty_from_duration(
    dtime, sigma, duration, confidence=0.95, distribution="normal"
):
    """
	*Give confidence interval length of a plan40 calculation.*

	:param dtime: Time step of your solutions
	:param sigma: Standard deviation of your signal
	:param duration: Desired duration of the signal
	:param confidence: Level of confidence of the interval
	:param distribution: Type of distribution of the signal to make the interval

	:returns:

	- **length** - Length of the confidence interval in K
	"""

    if distribution == "normal":
        z_coef = stats.norm.ppf(1 - (1 - confidence) / 2)
        # print(z_coef)
        n_timesteps = int(duration / dtime)
        length = 2 * z_coef * sigma / np.sqrt(n_timesteps)

    return length


def convergence_cartography(time, signal, **kwargs):
    """
	*Create a cartography of the convergence of the confidence interval in a simulation.*

	:param time: Time vector of your signal
	:param signal: Signal vector


	==**kwargs==
	
		:param max_time: Maximal simulation duration
		:param interlen: Maximal interval length

	:returns:
		- **fig** - Figure of the cartography

	"""
    max_time = kwargs.get("max_time", None)
    interlen = kwargs.get("interlen", None)

    if max_time is None:
        max_time = 0.05
    if interlen is None:
        interlen = 50

    dtime = calc_autocorrelation_time(time, signal)
    sigma = np.std(signal)
    durations = np.arange(0.001, max_time, 0.001)
    interval_lengths = np.arange(0, interlen)
    level = np.zeros((len(durations), len(interval_lengths)))
    for idx, dur in enumerate(durations):
        for idy, length in enumerate(interval_lengths):
            sqrtn = int(np.sqrt(dur / dtime))
            z_star = length * sqrtn / (2 * sigma)
            dec_per = stats.norm.cdf(z_star)
            level[idx, idy] = 1 - (1 - dec_per) * 2

    fig = plt.figure()
    fig = plt.contour(
        durations, interval_lengths, np.transpose(level), [0.6, 0.8, 0.9, 0.95, 0.975]
    )
    # plt.legend()
    plt.xlabel("Simulation duration (s)")
    plt.ylabel("Interval amplitude (K)")
    plt.clabel(
        fig,
        inline=1,
        fontsize=10,
        colors="k",
        fmt={0.6: "60%", 0.8: "80%", 0.9: "90%", 0.95: "95%", 0.975: "97.5%"},
    )
    plt.title("Contour of confidence interval(simulation duration, interval amplitude)")
    plt.grid(True)
    return fig


def calculate_std(time, signal, frequency):
    """
	*Give the standard deviation of a signal at a given frequency.*
	
	:param time: Time vector of your signal
	:param signal: Signal vector
	:param frequency: Frequency at which values of the recording are taken

	:returns:

		- **std** - Standard deviation of the values taken from the recording
	"""

    dtime = 1 / frequency
    time -= time[0]
    rescaled_time = np.arange(0, time[-1], dtime)
    rescaled_signal = np.interp(rescaled_time, time, signal)

    std = np.std(rescaled_signal)

    return std


def power_spectral_density(time, signal):
    """
	*Automate the computation of the Power Spectral Density of a signal.*

	:param time: Time vector of your signal
	:param signal: Signal vector

	:returns:

	- **frequency** -Frequency vector of the signal's power spectral density
	- **power_spectral_density** -Power spectral density of the signal
	"""

    total_time = time[-1] - time[0]
    dtime = time[1] - time[0]

    freq_resolution = 1.0 / total_time
    freq_nyquist = 1.0 / dtime

    N = len(time)

    frequency = np.arange(0, freq_nyquist, freq_resolution, dtype=np.float)
    frequency = frequency[: int(N / 2) - 1]

    mean_amplitude = np.mean(signal)

    raw_fft = np.fft.fft(signal - mean_amplitude)
    power_spectral_density = np.square(
        np.absolute(raw_fft)[: int(N / 2) - 1] / np.sqrt(N)
    )
    power_spectral_density[1:] *= 2

    return frequency, power_spectral_density


def to_percent(y, position):
    """
	*Rescale the y-axis to per*
    
    """
    s = str(round(100.0 * y))

    # The percent symbol needs escaping in latex
    if matplotlib.rcParams["text.usetex"] is True:
        return s + r"$\%$"
    else:
        return s + "%"


def plot_distributions(path="./data.txt"):
    formatter = FuncFormatter(to_percent)
    data = np.genfromtxt(path)
    S = np.shape(data)

    t_fusion = 1930.0
    bin_edges = list(np.linspace(t_fusion - 200.0, t_fusion + 200.0, 21))

    for j in range(S[1]):

        temperatures = data[:, j]

        clipped_temperatures = np.clip(temperatures, bin_edges[0], bin_edges[-1])
        clipped_temperatures.sort()
        weights = np.ones_like(clipped_temperatures) / float(len(clipped_temperatures))
        lower = 0
        upper = -1
        lower = np.argmax(clipped_temperatures > bin_edges[0]) - 1
        upper = np.argmax(clipped_temperatures == bin_edges[-1])

        if upper == 0:
            upper = -1

        fig, (ax1, ax2, ax3) = plt.subplots(
            1, 3, sharey=True, gridspec_kw={"width_ratios": [1, 5, 1]}
        )

        rect = plt.Rectangle((0.4, 0), 0.2, np.sum(weights[: lower + 1]))
        ax1.add_patch(rect)
        plt.sca(ax1)
        plt.xticks([0.5], ["<" + str(t_fusion - 200)])
        plt.gca().yaxis.set_major_formatter(formatter)
        ax1.set_ylabel("Probability")

        ax2.hist(
            clipped_temperatures[lower:upper],
            bins=bin_edges,
            weights=weights[lower:upper],
        )
        xticks = list(np.linspace(t_fusion - 200.0, t_fusion + 200.0, 11))
        xlabels = list(map(str, xticks))
        plt.sca(ax2)
        plt.xticks(xticks, xlabels, rotation=45, size=8)
        plt.gca().yaxis.set_major_formatter(formatter)

        rect = matplotlib.patches.Rectangle((0.4, 0), 0.2, np.sum(weights[upper:]))
        ax3.add_patch(rect)
        plt.sca(ax3)
        plt.xticks([0.5], [str(t_fusion + 200) + "<"])
        plt.gca().yaxis.set_major_formatter(formatter)

        plt.savefig("pdf_h_%s.png" % j)
        plt.close()



""" axishell module
to create x-axisymmetric shells for scientific computations
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json

import numpy as np
from scipy import interpolate, spatial

MESH_CTRL_POINTS = "axishell.json"
MAX_SPLINE_ORDER = 3
SPLINE_SMOOTHNESS = 0

MATRIX_ELEMENTS = ["xyz", "r", "theta",
                   "n_x", "n_y", "n_z", "n_r"]
AXIS_NAMES = ['time', 'theta', 'r']

class AxiShell():
    r"""\
    *Base class for x-axisymmetric computationnal shells*

    ::
    
                ___________
         ___----           ----___     Cylindrical system
        \                         /    Example angle = 0 deg
         \                       /
          \                     /      r - longi
           \                   /       ^
            \                 /        |
             \      ___      /         |
              \__---   ---__/          X-----> theta - azi
                                      x


    :param n_longi: Number of longitudinal cut points
    :type n_longi: int
    :param n_azi: Number of azimuthal cut points
    :type n_azi: int

    :param shape: Number of cut points n_longi and n_azi
    :type shape: tuple of int(dim 2)
    :param geom: dict() containing the geometrical parameters :

        - **angle**      - float - Axi-symmetric angle range
        - **angle_min**  - float - Minimum axi-symmetric angle
        - **ctrl_pts_x** - tuple of float - x-component of the spline control point
        - **ctrl_pts_r** - tuple of float - r-component of the spline control point
    :param matrix: dict() containing shell data :

        - **xyz**   - np.array of dim (n_azi,n_longi,3) - Array of x,y,z-components
        - **r**     - np.array of dim (n_azi,n_longi) - Array of r-component
        - **theta** - np.array of dim (n_azi,n_longi) - Array of theta-component
        - **n_x**   - np.array of dim (n_azi,n_longi) - Array of normal x-component
        - **n_y**   - np.array of dim (n_azi,n_longi) - Array of normal y-component
        - **n_z**   - np.array of dim (n_azi,n_longi) - Array of normal z-component
        - **n_r**   - np.array of dim (n_azi,n_longi) - Array of normal r-component

    :param cake: dict() containing 3D mesh from 2D shell extrusion :

        - *xyz* - np.array of dim (n_azi, n_longi, n_layers, 3) - Array of x,y,z-components
        - *dz*  - np.array of dim (n_azi, n_longi, n_layers) - Array of width per layer

    """
    def __init__(self, n_longi, n_azi):
        """
        *Initialize an AxiShell object*
        """

        self.shape = (n_azi, n_longi)
        # dict for the definition of the shell
        self.geom = {}

        # numpy array for the node based geometric elements
        self.matrix = {}

        self.cake = {}

    def init_mockup(self):
        """
        *Initialize with a mockup mesh*
        """
        self.geom["angle"] = 40.0

        # by nodes trapvtx
        self.geom["crtl_pts_x"] = (0.0,
                                   0.2)
        self.geom["crtl_pts_r"] = (0.12,
                                   0.12)

        # # sphere
        # self.geom["crtl_pts_x"] = 0.1 + 0.1* np.sin(
        #     np.linspace(-np.pi*0.5, np.pi*0.5, 4))
        # self.geom["crtl_pts_r"] = 0.1*np.cos(
        #     np.linspace(-np.pi*0.5, np.pi*0.5, 4))

        # by nodes curved shape
        # self.geom["crtl_pts_x"] = (0.0,
        #                            0.1,
        #                            0.2,
        #                            0.15)
        # self.geom["crtl_pts_r"] = (0.1,
        #                            0.15,
        #                            0.1,
        #                            0.03)

        self.build_shell()

    def load(self):
        """
        *Load AxiShell geometric features in JavaScript Object Notation*
        """
        with open(MESH_CTRL_POINTS, "r") as fin:
            self.geom = json.load(fin)

        self.build_shell()

    def dump(self):
        """
        *Dump AxiShell geometric features in JavaScript Object Notation*
        """

        with open(MESH_CTRL_POINTS, "w") as fout:
            json.dump(self.geom, fout)

    def build_shell(self):
        """
        *Build shell from geometric features*

            - Construct a spline used as base for extrusion\
              from control points : tck
            - Discretise the spline : shell_crest
            - Compute normal vectors for the 1D shell_crest
            - Compute r,n_x,n_r-components for 2D shell
            - Compute theta-components for 2D shell
            - Compute xyz,n_y,n_z-components for 2D shell
        """

        # Build crest
        spline_order = min(len(self.geom["crtl_pts_x"]) - 1, MAX_SPLINE_ORDER)
        # Generate continuous spline from control points.
        tck, _ = interpolate.splprep([self.geom["crtl_pts_x"],
                                      self.geom["crtl_pts_r"]],
                                     s=SPLINE_SMOOTHNESS,
                                     k=spline_order)
        unew = np.linspace(0,
                           1,
                           num=self.shape[1])
        # Generate discrete spline
        # Numpy Array of dim (2, self.shape[1]) [x,r]
        shell_crest = np.asarray(interpolate.splev(unew, tck))

        # Compute normal vectors
        normal_vects = np.roll(np.diff(shell_crest), 1, axis=0)
        normal_vects /= np.linalg.norm(normal_vects, axis=0)
        normal_vects[0, :] *= -1
        normal_vects2 = np.ones((2, self.shape[1]))
        normal_vects2[:, :self.shape[1] - 1] = normal_vects
        normal_vects2[:, -1] = normal_vects[:, -1]

        # Initialise each element
        for item in MATRIX_ELEMENTS:
            self.matrix[item] = np.ones(self.shape)

        # Compute r,n_x,n_r-components for 2D shell.
        tmp_x = np.tile(shell_crest[0], (self.shape[0], 1))
        self.matrix["r"] = np.tile(shell_crest[1], (self.shape[0], 1))
        self.matrix["n_x"] = np.tile(normal_vects2[0], (self.shape[0], 1))
        self.matrix["n_r"] = np.tile(normal_vects2[1], (self.shape[0], 1))

        ## Added to enable shell rotation and starting angle at
        ## value prescribed by user. This value is set by default to 0
        ## which corresponds to an axishell centered around y = 0.
        rot_angle = 0
        if "angle_min" in self.geom:
            rot_angle = (0.5 * self.geom["angle"] + self.geom["angle_min"])
        min_theta = (rot_angle - 0.5 * self.geom["angle"]) * np.pi / 180
        max_theta = (rot_angle + 0.5 * self.geom["angle"]) * np.pi / 180
        self.matrix["theta"] = np.transpose(
            np.tile(np.linspace(min_theta, max_theta, num=self.shape[0]),
                    (self.shape[1], 1)))

        tmp_y = (self.matrix["r"] * np.cos(self.matrix["theta"]))
        tmp_z = (self.matrix["r"] * np.sin(self.matrix["theta"]))

        self.matrix["xyz"] = np.stack((tmp_x, tmp_y, tmp_z),
                                      axis=2)

        self.matrix["n_y"] = (self.matrix["n_r"]
                              * np.cos(self.matrix["theta"]))
        self.matrix["n_z"] = (self.matrix["n_r"]
                              * np.sin(self.matrix["theta"]))

    def add_curviwidth(self, label, points):
        """
        *Add a 2D width matrix of shell shape extruded from points spline*

        :param label: Label of the width matrix
        :type label: str
        :param points: Tuple (dim n) of tuple (dim 2) of float coordinates\
        """

        (x_tuple, y_tuple) = tuple(zip(*points))

        # Extend Bounds
        xlist = [0] + list(x_tuple) + [1]
        ylist = [y_tuple[0]] + list(y_tuple) + [y_tuple[-1]]
        # Generate continuous fictive spline
        f_int = interpolate.interp1d(xlist, ylist)
        # Generate discretise spline
        xnew = np.linspace(0, 1, num=self.shape[1])
        ynew = f_int(xnew)

        # Extrude azimuthally
        self.matrix[label] = np.tile(ynew, (self.shape[0], 1))

    def set_mask_on_shell(self, point_cloud, tol):
        """
        *Create a mask on the shell from a point cloud*

        The mask value is 1 for shell points located near cloud points.

        :param point_cloud: Array of dim (n,3) of coordinates of points.
        :type point_cloud: numpy array
        :param tol: Tolerance of proximity
        :type tol: int
        """

        # Create a KDTree from cloud points
        kdtree = spatial.KDTree(point_cloud)
        # Compute distances between cloud points and
        # the closest neighbor from the 2D shell points.
        dists, _ = kdtree.query(self.matrix["xyz"], k=1)
        # Mask takes 1 if the distance is below tolerance.
        self.matrix["mask"] = np.where(
            dists.reshape(self.shape) > tol,
            np.zeros(self.shape),
            np.ones(self.shape))

    def bake_millefeuille(self, width_matrix_label, n_layers, shift=0.0):
        """
        *Create a millefeuille-like shell.*

        Extrude a 2D shell in the normal direction up
        pointwise height given by "width_matrix_label" matrix.

        :param width_matrix_label: Label of the width matrix
        :type width_matrix_label: str
        :param n_layers: Number of layer for extrusion
        :type n_layers: int
        :param shift: Additional depth (optional)
        :type shift: float

        :returns:

            - **cake** - A dict() containing shell data :

                - *xyz* - np.array of dim (n_longi, n_azi, n_layers, 3)
                - *dz*  - np.array of dim (n_longi, n_azi, n_layers)

        "Bon appetit!"
        """

        cake = {}
        cake["xyz"] = np.empty((self.shape[0],
                                self.shape[1],
                                n_layers,
                                3))
        cake["dz"] = np.empty((self.shape[0],
                               self.shape[1],
                               n_layers))

        for j, dim in enumerate(["x", "y", "z"]):
            for i in range(n_layers):
                cake["xyz"][:, :, i, j] = (self.matrix["xyz"][:, :, j]
                                           + (1.0 * i / n_layers
                                              * self.matrix[width_matrix_label]
                                              + shift)
                                           * self.matrix["n_" + dim])

        for i in range(n_layers):
            cake["dz"][:, :, i] = abs(self.matrix[width_matrix_label] / n_layers)
        cake["dz"][:, :, 0] /= 2
        cake["dz"][:, :, -1] /= 2

        return cake

    def average_on_shell_over_dirs(self, variable, directions):
        """
        *Performs an integration (averaging) over one or multiple directions*

        :param variable: A np.array to be averaged of dim (n_time, n_theta, n_r)
        :param directions: A list() of directions on which the average process\
                           is to be performed.\
                           Contains keywords from ['time','theta','r'].

        :returns:

            - **averaged_variable** - A np.array of averaged data on given directions.
        """

        scaled = np.copy(variable)
        if self.matrix.get("surf") is not None:
            scaled = np.multiply(variable, self.matrix.get("surf"))

        dirs = AXIS_NAMES.copy()

        if (len(directions) > 0 and
                all(dir_ in dirs for dir_ in directions)):
            averaged_variable = scaled
            for dir_ in directions:
                index = dirs.index(dir_)
                averaged_variable = np.mean(averaged_variable, axis=index)
                dirs.pop(index)

        else:
            mess = 'Averaging directions do not conform to criteria\n'
            mess += 'It should be a list containing one of or all items\n'
            mess += '"time", "theta", "r"'
            raise ValueError(mess)

        return averaged_variable

def width_mockup():
    """ Create a mockup tuple of tuple for widths"""

    width_mockup_ = ((0.0, -0.04),
                     (0.2, -0.04),
                     (0.3, -0.07),
                     (0.4, -0.07),
                     (0.5, -0.04),
                     (0.8, -0.04))

    return width_mockup_




"""
======================
Solver 2D (Deprecated)
======================

This package contains a solver of second partial derivative equations to treat heat conduction and heat radiation problem.
The 2nd order finit difference scheme is used to solve the inside of a computational domain and that of first order for boundaries.

This development is now hosted in a separate repository at `nitrox - calcifer <https://nitrox.cerfacs.fr/open-source/calcifer>`_
"""

from .boundary import *
from .conduction import *
from .core_fd import *
from .radiation import *



""" module to compute wall equilibrium """
import numpy as np

__all__ = ["compute_equilibrium"]


def compute_equilibrium(
        hot_t_ad,
        cold_t_ad,
        hot_h,
        cold_h,
        metal,
        ceram,
        ep_metal,
        ep_ceram):
    """ 
    compute the wall equilibrium
    -----------------------------

    ::

                - - - - - - - >
                HOT SIDE  hot_h, hot_t_ad

             ^ phi
            _|___________________  t_ceram_hot
             |       Layer 2
            _|___________________ t_eq
             |       Layer 1
            _|___________________ t_metal_hot
             |
                COLD SIDE cold_h, cold_t_ad
                - - - - >
    """

    t_eq = np.copy(cold_t_ad)
    t_eq_m1 = np.copy(cold_t_ad)

    def estimate_phi_and_temps(t_ceram, t_metal):
        """ return phi and temperatures """
        r_th_ceram = ceram.thermal_resistance(ep_ceram, t_ceram)
        r_th_metal = metal.thermal_resistance(ep_metal, t_metal)

        phi = (hot_t_ad - cold_t_ad) / (1.0 / hot_h +
                                        r_th_ceram +
                                        r_th_metal +
                                        1.0 / cold_h)

        t_ceram_hot = hot_t_ad - phi / hot_h
        t_ceram_metal = t_ceram_hot - phi * r_th_ceram
        t_metal_cold = t_ceram_metal - phi * r_th_metal

        return t_ceram_hot, t_ceram_metal, t_metal_cold

    t_ceram_hot, t_eq, t_metal_cold = estimate_phi_and_temps(t_ceram=t_eq,
                                                             t_metal=t_eq)

    # Stop convergence at 1K
    ncount = 0
    cvgce = 1000.
    while cvgce > 1.0:
        ncount += 1
        t_eq_m1 = t_eq
        t_ceram_hot, t_eq, t_metal_cold = estimate_phi_and_temps(
            t_ceram = 0.5 * (t_ceram_hot + t_eq),
            t_metal = 0.5 * (t_metal_cold + t_eq))
        cvgce =  np.max(np.abs(t_eq - t_eq_m1)) 

        if ncount > 30:
            msgerr = "Compute_equilibrium loop did not conveg."
            msgerr = "Breaking with a deltaT of  " + str(cvgce) + "K"
            raise RuntimeWarning(msgerr)

    return t_ceram_hot, t_eq, t_metal_cold



"""Module for computing thermodynamic properties
"""

import numpy as np


__all__ = [
    "thermal_constants",
    "h_kader",
    "lambda_cp_visco_fluid",
    "viscosity_sutherland",
    "fluid_cp"
]

def thermal_constants():
    """Generate a dictionnary of thermal constants

    Returns:
        TYPE: Description
    """
    consts = dict()
    consts['prandtl'] = 0.71
    consts['prandtl_turb'] = 0.86
    consts['vkarman'] = 0.41
    consts['van_driest'] = 5.5
    consts['beta_kader'] = ((3.85 * consts['prandtl']**(1.0/3.0) - 1.3)**2
                            + 2.12 * np.log(consts['prandtl']))
    consts['k_kader'] = ((consts['beta_kader']
                          - consts['prandtl_turb'] * consts['van_driest']
                          + consts['prandtl_turb'] / consts['vkarman'] - 2.12)
                         - (consts['prandtl_turb'] / consts['vkarman'] - 2.12)
                         * (200.0*np.log(200.0) - 100*np.log(100.0))
                         / (200.0 - 100.0))
    return consts


def _estimate_log_region(y_plus_guess, u_2, y_rho_mu_wall):
    """evaluate u tau in log region

    Returns:
        TYPE: Description

    Args:
        y_plus_guess (TYPE): Description
        u_2 (TYPE): Description
        y_rho_mu_wall (TYPE): Description
    """
    u_tau_log = 0
    epsilon = 1.0
    y_plus_log = y_plus_guess
    consts = thermal_constants()

    def _loop_u_tau(u_tau_log, y_plus_log):
        """recursive loop to evaluate u_tau

        Args:
            u_tau_log (TYPE): Description
            y_plus_log (TYPE): Description

        Returns:
            TYPE: Description

        """
        utau_old = u_tau_log
        u_tau_log = u_2 / (1 / consts['vkarman'] * np.log(y_plus_log)
                           + consts['van_driest'])
        y_plus_log = u_tau_log * y_rho_mu_wall
        epsilon = np.abs(u_tau_log - utau_old)
        return u_tau_log, y_plus_log, epsilon

    while np.min(epsilon) >= 1.e-5:
        (u_tau_log, y_plus_log, epsilon) = _loop_u_tau(u_tau_log, y_plus_log)
    return u_tau_log, y_plus_log

def _estimate_t_tau(u_tau, y_plus, u_2, t_wall, t_2):
    """Compute T_tau

    Args:
        u_tau (TYPE): Description
        y_plus (TYPE): Description
        u_2 (TYPE): Description
        t_wall (TYPE): Description
        t_2 (TYPE): Description

    Returns:
        TYPE: Description
    """
    consts = thermal_constants()
    kader_g = (0.01 * (consts['prandtl'] * y_plus['all'])**4)
    kader_g /= (1.0 + 5.0 * consts['prandtl']**3 * y_plus['all'])
    t_tau_cwm = (
        (t_wall - t_2)
        / ((consts['prandtl'] * y_plus['all'] * np.exp(-kader_g))
           + (consts['prandtl_turb']*u_2/u_tau['cwm'] +  consts['k_kader'])
           * np.exp(-1.0/kader_g)))
    return t_tau_cwm

def h_kader(t_wall, rho_wall, y_wall, u_2, t_2, temp_adiab):
    """compute h at the wall as in kader
        names taken equalt to  loglaw_cwm.f90 AVBP

    Args:
        t_wall (TYPE): Description
        rho_wall (TYPE): Description
        y_wall (TYPE): Description
        u_2 (TYPE): Description
        t_2 (TYPE): Description
        temp_adiab (TYPE): Description

    Returns:
        TYPE: Description
    """
    _, heat_cp, mu_wall = lambda_cp_visco_fluid(t_2)

    y_plus = dict()
    y_rho_mu_wall = y_wall * rho_wall / mu_wall
    y_plus['guess'] = 0.157 * (u_2 * y_rho_mu_wall)**(7.0/8.0)
    y_plus['guess'] = np.maximum(y_plus['guess'], 1.0e-3)

    u_tau = dict()
    # Linear part
    u_tau['lin'] = np.sqrt(u_2 / y_rho_mu_wall)

    #Log part
    u_tau['log'], y_plus['log'] = _estimate_log_region(y_plus['guess'], u_2,
                                                       y_rho_mu_wall)

    # fusion of the two
    u_tau['cwm'] = np.where(y_plus['guess'] < 11.25, u_tau['lin'], u_tau['log'])
    y_plus['all'] = u_tau['cwm'] * y_rho_mu_wall

    t_tau_cwm = _estimate_t_tau(u_tau, y_plus, u_2, t_wall, t_2)
    h_wall = -(rho_wall * heat_cp * u_tau['cwm'] * t_tau_cwm)
    h_wall /= (temp_adiab - t_wall)

    return h_wall


def lambda_cp_visco_fluid(temperature):
    """compute Fluid properties lambda , cp, visco

    Args:
        temperature (TYPE): Description

    Returns:
        TYPE: Description
    """
    consts = thermal_constants()
    heat_cp = fluid_cp(temperature)
    mu_wall = viscosity_sutherland(temperature)
    lam = heat_cp / consts['prandtl'] * mu_wall

    return lam, heat_cp, mu_wall

def viscosity_sutherland(temp):
    """compute visocity as in sutherland

    Args:
        temp (TYPE): Description

    Returns:
        TYPE: Description

    """
    alpha = 1.716e-5
    beta = 1.5
    t_0 = 273.15
    t_1 = 110.4
    mu_wall = alpha * (temp / t_0)**beta * (t_0 + t_1) / (temp + t_1)
    return mu_wall

def fluid_cp(temp, clipping=False):
    """compute cp of fluid

    Args:
        temp (TYPE): Description
        clipping (bool, optional): Description

    Returns:
        TYPE: Description

    """
    t_clip = np.array(temp)
    if clipping:
        t_clip = np.clip(t_clip, 300.0, 2500.0)
    coeffs = [513.46, 1.6837, -1.9275e-3, 1.2773e-6, -4.2773e-10, 5.6735e-14]
    heat_cp = (
        coeffs[0] * np.ones_like(t_clip)
        + coeffs[1] * t_clip
        + coeffs[2] * np.power(t_clip, 2)
        + coeffs[3] * np.power(t_clip, 3)
        + coeffs[4] * np.power(t_clip, 4)
        + coeffs[5] * np.power(t_clip, 5)
    )
    return heat_cp



""" Module concerning some 3D vector manipulations in numpy

OST :Mercy in Darkness, by Two Steps From Hell

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from scipy.spatial.transform import Rotation as Rot

SMALL = 1e-12


def renormalize(np_ar_vect):
    """ renormalize a numpy array of vectors considering the last axis
    """
    result = (np_ar_vect
              / np.expand_dims(np.linalg.norm(np_ar_vect, axis=-1),
                               axis=-1))
    return result


def angle_btw_vects(np_ar_vect1, np_ar_vect2, convert_to_degree=False):
    """ compute the angle in deg btw two UNIT vectors """
    sdot = np.sum(renormalize(np_ar_vect1)
                  * renormalize(np_ar_vect2), axis=-1)

    # Clipping is needed when pscal.max() > 1 : 1.000000000002
    sdot = np.clip(sdot, -1.0, 1.0)
    angles = np.arcsin(sdot)
    if convert_to_degree:
        angles = np.rad2deg(angles)
    return angles


def yz_to_theta(np_ar_vect):
    """ return theta , a radians measure of ange in the yz plane,
    - range -pi/pi
    - 0 on the y+ axis (z=0, y>0)
    spanning -pi to pi
                   0pi=0deg
                  Y
                  ^
                  |
                  |
-0.5pi=-90deg     o------>Z   0.5pi=90deg

    """
    # return np.arctan2(np_ar_vect[:, 2],
    #                   np_ar_vect[:, 1])
    return np.arctan2(np.take(np_ar_vect, 2, axis=-1),
                      np.take(np_ar_vect, 1, axis=-1))


def rtheta2yz(rrr, theta):
    """ return yz fror rtheta ,
    theta in  radians measure of ange in the yz plane,
    - range -pi/pi
    - 0 on the y+ axis (z=0, y>0)
    spanning -pi to pi

::

                        0pi=0deg
                      Y
                      ^
                      |
                      |
    -0.5pi=-90deg     o------>Z   0.5pi=90deg

    """
    yyy = rrr * np.cos(theta)
    zzz = rrr * np.sin(theta)
    return yyy, zzz


def rotate_vect_around_x(np_ar_vect, angle_deg):
    """ rotate vector around axis x in degree """

    arr_r = np.hypot(np.take(np_ar_vect, 1, axis=-1),
                     np.take(np_ar_vect, 2, axis=-1))
    arr_theta = (yz_to_theta(np_ar_vect)
                 + np.deg2rad(angle_deg))

    result = np.stack((np.take(np_ar_vect, 0, axis=-1),
                       arr_r * np.cos(arr_theta),
                       arr_r * np.sin(arr_theta)),
                      axis=-1)

    return result

def rotate_vect_around_axis(xyz, *tuples_rot):
    """
    *Rotate vector around vector or series of vector*

    :param xyz: Array of xyz-coordinates of shape (n,3)
    :param tuples_rot: List of tuple with rotation data : \
                       Axis array of shape (3,) axis, \
                       Float angle in degree

    :returns: Array of rotated xyz-coordinates of shape (n,3)
    """
    quat = Rot.from_rotvec([0, 0, 0])
    for axis, angle in tuples_rot:
        rot_vect = np.multiply(renormalize(axis),
                               angle * np.pi / 180.)
        quat *= Rot.from_rotvec(rot_vect)

    return quat.apply(xyz)

def dilate_vect_around_x(azimuth, np_ar_vect, angle_deg_init=None, angle_deg_targ=360):
    """ dilate vectors around axis x from a specified initial range angle
        to a target range angle.
    Parameters :
    ------------
    np_ar_vect : numpy array of dim (n,3)
    angle_deg_targ : tuple or float

    Returns :
    ---------
    numpy array of dim (n,3)
    """

    if angle_deg_init is None:
        angle_deg_init = np.max(azimuth) - np.min(azimuth)
    dilate_factor = angle_deg_targ / angle_deg_init

    result = rotate_vect_around_x(np_ar_vect,
                                  azimuth * (dilate_factor - 1))
    return result


def cart_to_cyl(vects_xyz):
    """
    *Transform vects from xyz-system to xrtheta-system*

    x -> x     : x = x
    y -> r     : r = sqrt(y^2 + z^2)
    z -> theta : theta = arctan2(z,y)

    :param vects_xyz: Array of dim (n,3) of xyz components
    :type vects_xyz: np.array

    :return:

        - **vects_cyl** - Array of dim (n,3) of xrtheta components
    """

    np_x = np.take(vects_xyz, 0, -1)
    np_r = np.sqrt(np.take(vects_xyz, 1, -1) ** 2
                   + np.take(vects_xyz, 2, -1) ** 2)
    np_theta = yz_to_theta(vects_xyz)

    vects_cyl = np.stack((np_x, np_r, np_theta), axis=-1)

    return vects_cyl

def clip_by_bounds(points_coord, bounds_dict, keep="in", return_mask=False):
    """
    *Clip a cloud by keeping only or removing a bounded region*

    The dict to provide must be filled as follow :
    bounds_dict = {component_1 : (1_min, 1_max),
                   compoment_2 : (2_min, 2_max),
                   ...}

    component_1 = ["x", "y", "z", "r", "theta"]

    The bounded region can either be :

        - A 1D slice if only 1 component is provided ;
        - A 2D box if 2 components are provided ;
        - A 3D box if 3 components are provided.

    If keep="in", returns the point coordinates inside the bounds.
    If keep="out", returns the point coordinates outside the bounds.

    If returns=True, returns the coordinates clipped
    If returns=False, returns the mask of boolean than can be applied\
    on other arrays

    :param point_cloud: Array of dim (n,k) of coordinates
    :type point_cloud: np.array
    :param bounds_dict: Dict of MAX lengh k of tuple of floats
    :param keep: Either keeps what is inside or outside
    :type keep: str

    :returns:

        - **points_coord_clipped** - Array of dim (m,k) with m<=n
        OR
        - **mask** - Array of dim (n,) of booleans
    """
    mask = np.ones(points_coord.shape[0], dtype=bool)
    for dim, bounds in bounds_dict.items():
        mask &= mask_cloud(points_coord, dim, bounds)

    if keep == "out":
        mask = np.invert(mask)

    if return_mask:
        return mask
    return points_coord[mask]

def cyl_to_cart(vects_cyl):
    """
    *Transform vects from xrtheta-system to xyz-system*

    x -> x     : x = x
    r -> y     : y = r * cos(theta)
    theta -> z : z = r * sin(theta)

    :param vects_cyl: Array of dim (n,3) of xrtheta components
    :type vects_cyl: np.array

    :return:

        - **vects_xyz** - Array of dim (n,3) of xyz components
    """

    np_x = np.take(vects_cyl, 0, -1)
    np_y = np.take(vects_cyl, 1, -1) * np.cos(np.take(vects_cyl, 2, -1))
    np_z = np.take(vects_cyl, 1, -1) * np.sin(np.take(vects_cyl, 2, -1))

    vects_xyz = np.stack((np_x, np_y, np_z), axis=-1)

    return vects_xyz

def vect_to_quat(vect_targ, vect_source):
    """
    *Generate a quaternion from two vectors*

    A quaternion is a rotation object. From two vectors,\
    the rotation angle and the rotation axis are computed.
    The rotation vector generates then a quaternion for each
    serie of vectors.

    :param vect_targ: Array of dim (n,3) of vect components
    :type vect_targ: np.array
    :param vect_source: Array of dim (n,3) of vect components
    :type vect_source: np.array

    :return:

        - **quat** - Array of quaternion of dim (n,)
    """
    vect_targ = renormalize(vect_targ)
    vect_source = renormalize(vect_source)

    prod_scal = np.clip(np.sum(vect_targ * vect_source,
                               axis=-1),
                        -1, 1)

    rot_ang = np.arccos(prod_scal)
    if vect_targ.ndim >= 2:
        rot_ang = rot_ang[:, np.newaxis]
    else:
        rot_ang = np.array([rot_ang])

    rot_axe = np.cross(vect_source, vect_targ)
    idx = np.where(rot_ang > 1e-6)[0]
    rot_axe[idx] = renormalize(rot_axe[idx])

    rot_vect = np.multiply(rot_ang, rot_axe)

    quat = Rot.from_rotvec(rot_vect)

    return quat


def make_radial_vect(coord, vects):
    """
    *Recalibrate vectors to make them radial.*

    The vectors are readjusted to cross x-axis. It is mainly done for nodes
    on the limit of the boundary for axi-cylindrical geometries.

    :param coord: Array of dim (n,3) of float coordinates
    :type coord: np.array
    :param vects: Array of dim (n,3) of float components
    :type vects: np.array

    :return:

        - **radial_vect** - Array of dim (n,3) of float components
    """

    tol = 10
    diff_angle = (yz_to_theta(coord) - yz_to_theta(vects)) * 180. / np.pi
    diff_angle = np.where(diff_angle > 90., diff_angle - 180., diff_angle)
    diff_angle = np.where(diff_angle < -90., diff_angle + 180., diff_angle)
    diff_angle = np.where(np.absolute(diff_angle) > tol, 0.0, diff_angle)

    radial_vect = rotate_vect_around_x(vects, diff_angle)
    return radial_vect

def mask_cloud(np_ar_xyz,
               axis,
               support):
    """ mask a cloud of n 3D points in in xyz
    axis among x,y,z,theta,r
    support a 2 value tuple : (0,3), (-12,float(inf))

    x and (0,3) reads as  0 <= x < 3 ( lower bound inclusive)
    z and (-12,float(inf)) reads as  -12 <= z
    theta in degree, cyl. coordinate around x axis
    - range -180,180
    - 0 on the y+ axis (z=0, y>0)
    """

    def true_between(np_arr, low, high):
        """ return a boolean array of the same shape,
        true if value are btw low and high"""
        return (np_arr >= low) * (np_arr <= high)

    possible_axes = ["x", "y", "z", "theta", "r"]
    if axis not in possible_axes:
        raise IOError("axis " + axis + "not among possible axes "
                      + ";".join(possible_axes))

    if axis == "x":
        mask = true_between(
            np.take(np_ar_xyz, 0, axis=-1),
            support[0],
            support[1])
    if axis == "y":
        mask = true_between(
            np.take(np_ar_xyz, 1, axis=-1),
            support[0],
            support[1])
    if axis == "z":
        mask = true_between(
            np.take(np_ar_xyz, 2, axis=-1),
            support[0],
            support[1])
    if axis == "r":
        mask = true_between(
            np.hypot(np.take(np_ar_xyz, 1, axis=-1),
                     np.take(np_ar_xyz, 2, axis=-1)),
            support[0],
            support[1])
    if axis == "theta":
        mask = true_between(
            np.rad2deg(yz_to_theta(np_ar_xyz)),
            support[0],
            support[1])
    return mask



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from glob import glob
from os.path import basename
#from os.path import dirname
#from os.path import join
from os.path import splitext
from setuptools import find_packages, setup

NAME = "arnica"
VERSION = "1.7.2"

setup(
    name=NAME,
    version=VERSION,
    description='Open Source library CFD toolkit',
    author='CoopTeam-CERFACS',
    author_email='coop@cerfacs.com',
    url='',
    keywords=["ARNICA"],
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    license="CeCILL-B FREE SOFTWARE LICENSE AGREEMENT",
    install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'mplcursors',
        'h5py',
        'PyYAML>=3.13',
        'lxml',
        'pandas',
        'hdfdict',
        ]
)



"""
ARNICA
======

Provides a python wrapper to setup, run and post-process AVBP simulations.

Available subpackages
---------------------
utils
    A collection of tools for geometrical problems
phys
    A collection of tools for physical problems, essentially around CHT configurations
solvers_2d
    Small 2d structured FD solvers in(Dperecated)
"""

from arnica import utils
from arnica import phys
from arnica import solvers_2d



""" axishell module
to create x-axisymmetric shells for scientific computations
"""

import numpy as np
from arnica.utils.shell import Shell

class AxiShell(Shell):
    #pylint: disable=too-many-instance-attributes
    r"""\
    *Base class for x-axisymmetric computationnal shells*

    AxiShell class is based on Shell class, inheriting of Shell methods.
    At initialization, the shell is builded an an x-axisymmetric shell.

    Some attributes are based on **u** and **v**, defined as the curvilinear\
    longitudinal (x/r) abscissa and curvilinear azimutal (theta) abscissa\
    respectively.

    ::
                      (x_n, r_n)
                _____X_____
         ___----     |     ----___     Cylindrical system
        \            |            /    Example longi/u <=> r
         \           |           /
          \          |          /      r - longi/u
           \         |         /       ^
            \        | (x_0, r_0)      |
             \      _X_      /         |
              \__---   ---__/          X-----> theta - azi/v
                                      x

              ................
          ...                 ..       Cylindrical system
        ..                     .       Example longi/u <=> x/r
        .                      .
        .      ................         r
        .     .                         ^
         .     ...................      |
         ..                             |
            ......................      X-----> x
                          <--      theta - azi/v
                     x/r - longi/u

    :param n_azi: Number of azimuthal shell points
    :type n_azi: int
    :param n_longi: Number of longitudinal shell points
    :type n_longi: int
    :param angle: Range angle of the axicylindrical geometry
    :type angle: float
    :param ctrl_pts_x: Array of dim (n,) of x-coordinates of the points defining the spline
    :param ctrl_pts_r: Array of dim (n,) of r-coordinates of the points defining the spline

    Optionals arguments :
    :param angle_min: Minimum angle of the shell
    :type angle_min: float

    Public attributes :

        - **shape** - Shape of the shell : (n_azi, n_longi)
        - **width_matrix** - Dict containing thickness matrices of shape 'shape'
        - **matrix** - Dict containing fields of shape 'shape' - To deprecate

    Private attributes :

        - **_xyz**
        - **_rad**
        - **_theta**
        - **_n_x**
        - **_n_r**
        - **_n_y**
        - **_n_z**
        - **_du**
        - **_dv**
        - **_abs_curv**
        - **_du**
        - **_dv**
        - **_abs_curv**
        - **_dwu**
        - **_dwv**
        - **_surf**

    """

    def __init__(self, n_azi, n_longi, angle, ctrl_pts_x, ctrl_pts_r, angle_min=None):
        #pylint: disable=too-many-arguments
        """
        *Initialize an AxiShell object*
        """

        Shell.__init__(self, n_azi, n_longi)

        self.ctrl_pts_x = ctrl_pts_x
        self.ctrl_pts_r = ctrl_pts_r
        self.angle = angle
        self.angle_min = angle_min

        self._build_shell()

    def _build_shell(self):
        #pylint: disable=arguments-differ
        """
        *Build shell from geometric features*

            - Construct a spline used as base for extrusion\
              from control points : tck
            - Discretise the spline : shell_crest
            - Compute normal vectors for the 1D shell_crest
            - Compute r,n_x,n_r-components for 2D shell
            - Compute theta-components for 2D shell
            - Compute xyz,n_y,n_z-components for 2D shell
        """

        # Construct Shell Crest
        shell_crest = self._compute_shell_crest(self.ctrl_pts_x, self.ctrl_pts_r)

        # Compute radius matrix of shape (n_azi, n_longi)
        self._rad = np.tile(shell_crest[1], (self.shape[0], 1))

        # Compute theta matrix of shape (n_azi, n_longi)
        rot_angle = 0
        if self.angle_min is not None:
            rot_angle = 0.5 * self.angle + self.angle_min
        min_theta = (rot_angle - 0.5 * self.angle) * np.pi / 180
        max_theta = (rot_angle + 0.5 * self.angle) * np.pi / 180
        self._theta = np.transpose(np.tile(np.linspace(min_theta,
                                                       max_theta,
                                                       num=self.shape[0]),
                                           (self.shape[1], 1)))

        # Compute xyz matrix of shape (n_azi, n_longi, 3)
        tmp_x = np.tile(shell_crest[0], (self.shape[0], 1))
        tmp_y = self._rad * np.cos(self._theta)
        tmp_z = self._rad * np.sin(self._theta)
        self._xyz = np.stack((tmp_x, tmp_y, tmp_z), axis=-1)

        # Compute x,y,z,r-normal matrix of shape (n_azi, n_longi)
        xr_nml_1d = self._compute_shellcrest_nml(shell_crest)
        self._n_r = np.tile(xr_nml_1d[1], (self.shape[0], 1))
        self._n_x = np.tile(xr_nml_1d[0], (self.shape[0], 1))
        self._n_y = self._n_r * np.cos(self._theta)
        self._n_z = self._n_r * np.sin(self._theta)

        # Compute du,dv matrix of shape(n_azi, n_longi)
        self._du = np.pad(np.sqrt(np.diff(self._xyz[:, :, 0], axis=1) ** 2
                                  + np.diff(self._rad, axis=1) ** 2),
                          ((0, 0), (1, 0)),
                          'edge')
        self._dv = self._rad * np.pad(np.diff(self._theta, axis=0),
                                      ((1, 0), (0, 0)),
                                      'edge')
        # Compute abs_curv array of shape (n_longi,)
        self._abs_curv = np.cumsum(np.take(self._du, 0, 0)) - self._du[0, 0]

        # Compute weight intervals in u and v directions array of shape (n_transvers, n_longi)
        self._dwu = self._du.copy()
        self._dwu[:, (0, -1)] /= 2
        self._dwv = self._dv.copy()
        self._dwv[(0, -1), :] /= 2

        # Compute surface nodoes array of shape (n_transvers, n_longi)
        self._surf = np.multiply(self._dwu, self._dwv)
